# Copyright (C) 2020 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from io import StringIO
import os
import sys
from unittest import mock

import ddt
from oslo_utils.fixture import uuidsentinel

from tackerclient import client as root_client
from tackerclient.common import exceptions
from tackerclient.osc import utils as tacker_osc_utils
from tackerclient.osc.v1.vnflcm import vnflcm
from tackerclient.tests.unit.osc import base
from tackerclient.tests.unit.osc.v1.fixture_data import client
from tackerclient.tests.unit.osc.v1 import vnflcm_fakes
from tackerclient.tests.unit.test_cli10 import MyResp
from tackerclient.v1_0 import client as proxy_client


class TestVnfLcm(base.FixturedTestCase):
    client_fixture_class = client.ClientFixture

    def setUp(self):
        super(TestVnfLcm, self).setUp()
        self.url = client.TACKER_URL
        self.header = {'content-type': 'application/json'}
        self.app = mock.Mock()
        self.app_args = mock.Mock()
        self.client_manager = self.cs
        self.app.client_manager.tackerclient = self.client_manager


def _get_columns_vnflcm(action='create'):
    columns = ['ID', 'Instantiation State', 'VNF Instance Description',
               'VNF Instance Name', 'VNF Product Name', 'VNF Provider',
               'VNF Software Version', 'VNFD ID', 'VNFD Version', 'Links',
               'VNF Configurable Properties', 'VNF Package ID']
    if action == 'show':
        columns.extend(['Instantiated Vnf Info', 'VIM Connection Info'])
    if action == 'list':
        columns = [ele for ele in columns if ele not in
                   ['VNFD Version', 'VNF Instance Description',
                    'VNF Configurable Properties', 'VNF Package ID']]
        columns.remove('Links')
    return columns


@ddt.ddt
class TestCreateVnfLcm(TestVnfLcm):

    def setUp(self):
        super(TestCreateVnfLcm, self).setUp()
        self.create_vnf_lcm = vnflcm.CreateVnfLcm(
            self.app, self.app_args, cmd_name='vnflcm create')

    def test_create_no_args(self):
        self.assertRaises(base.ParserException, self.check_parser,
                          self.create_vnf_lcm, [], [])

    @ddt.data({"optional_arguments": True, "instantiate": True},
              {"optional_arguments": False, "instantiate": False})
    @ddt.unpack
    def test_take_action(self, optional_arguments, instantiate):
        arglist = [uuidsentinel.vnf_package_vnfd_id]
        verifylist = [('vnfd_id', uuidsentinel.vnf_package_vnfd_id)]

        if optional_arguments:
            arglist.extend(['--name', 'test',
                            '--description', 'test'])
            verifylist.extend([('name', 'test'),
                               ('description', 'test')])

        # command param
        if instantiate:
            param_file = ("./tackerclient/osc/v1/vnflcm/samples/"
                          "instantiate_vnf_instance_param_sample.json")

            arglist.extend(['--I', param_file])
            verifylist.append(('I', param_file))

        parsed_args = self.check_parser(self.create_vnf_lcm, arglist,
                                        verifylist)

        json = vnflcm_fakes.vnf_instance_response()
        self.requests_mock.register_uri(
            'POST', os.path.join(self.url, 'vnflcm/v1/vnf_instances'),
            json=json, headers=self.header)

        if instantiate:
            self.requests_mock.register_uri(
                'POST', os.path.join(self.url, 'vnflcm/v1/vnf_instances',
                                     json['id'], 'instantiate'),
                json={}, headers=self.header)

        sys.stdout = buffer = StringIO()
        actual_columns, data = (self.create_vnf_lcm.take_action(parsed_args))

        headers, attributes = vnflcm._get_columns(json)

        expected_message = (
            'VNF Instance ' + json['id'] + ' is created and instantiation '
                                           'request has been accepted.')
        if instantiate:
            self.assertEqual(expected_message, buffer.getvalue().strip())

        self.assertCountEqual(_get_columns_vnflcm(),
                              actual_columns)
        self.assertListItemsEqual(vnflcm_fakes.get_vnflcm_data(
            json, columns=attributes), data)


class TestShowVnfLcm(TestVnfLcm):

    def setUp(self):
        super(TestShowVnfLcm, self).setUp()
        self.show_vnf_lcm = vnflcm.ShowVnfLcm(
            self.app, self.app_args, cmd_name='vnflcm show')

    def test_take_action(self):
        vnf_instance = vnflcm_fakes.vnf_instance_response(
            instantiation_state='INSTANTIATED')

        arglist = [vnf_instance['id']]
        verifylist = [('vnf_instance', vnf_instance['id'])]

        # command param
        parsed_args = self.check_parser(self.show_vnf_lcm, arglist,
                                        verifylist)

        self.requests_mock.register_uri(
            'GET', os.path.join(self.url, 'vnflcm/v1/vnf_instances',
                                vnf_instance['id']),
            json=vnf_instance, headers=self.header)

        columns, data = (self.show_vnf_lcm.take_action(parsed_args))
        self.assertCountEqual(_get_columns_vnflcm(action='show'),
                              columns)
        headers, attributes = vnflcm._get_columns(vnf_instance, action='show')
        self.assertListItemsEqual(
            vnflcm_fakes.get_vnflcm_data(vnf_instance, columns=attributes),
            data)


class TestListVnfLcm(TestVnfLcm):

    def setUp(self):
        super(TestListVnfLcm, self).setUp()
        self.list_vnf_instance = vnflcm.ListVnfLcm(
            self.app, self.app_args, cmd_name='vnflcm list')

    def test_take_action(self):
        vnf_instances = vnflcm_fakes.create_vnf_instances(count=3)
        parsed_args = self.check_parser(self.list_vnf_instance, [], [])
        self.requests_mock.register_uri(
            'GET', os.path.join(self.url, 'vnflcm/v1/vnf_instances'),
            json=vnf_instances, headers=self.header)
        actual_columns, data = self.list_vnf_instance.take_action(parsed_args)

        headers, columns = tacker_osc_utils.get_column_definitions(
            vnflcm._attr_map, long_listing=True)

        expected_data = []
        for vnf_instance_obj in vnf_instances:
            expected_data.append(vnflcm_fakes.get_vnflcm_data(
                vnf_instance_obj, columns=columns, list_action=True))

        self.assertCountEqual(_get_columns_vnflcm(action='list'),
                              actual_columns)
        self.assertCountEqual(expected_data, list(data))

    def test_take_action_with_pagination(self):
        vnf_instances = vnflcm_fakes.create_vnf_instances(count=3)
        next_links_num = 3
        parsed_args = self.check_parser(self.list_vnf_instance, [], [])
        path = os.path.join(self.url, 'vnflcm/v1/vnf_instances')

        links = [0] * next_links_num
        link_headers = [0] * next_links_num

        for i in range(next_links_num):
            links[i] = (
                '{base_url}?nextpage_opaque_marker={vnf_instance_id}'.format(
                    base_url=path,
                    vnf_instance_id=vnf_instances[i]['id']))
            link_headers[i] = copy.deepcopy(self.header)
            link_headers[i]['Link'] = '<{link_url}>; rel="next"'.format(
                link_url=links[i])

        self.requests_mock.register_uri(
            'GET', path, json=[vnf_instances[0]], headers=link_headers[0])
        self.requests_mock.register_uri(
            'GET', links[0], json=[vnf_instances[1]],
            headers=link_headers[1])
        self.requests_mock.register_uri(
            'GET', links[1], json=[vnf_instances[2]],
            headers=link_headers[2])
        self.requests_mock.register_uri(
            'GET', links[2], json=[], headers=self.header)

        actual_columns, data = self.list_vnf_instance.take_action(parsed_args)

        headers, columns = tacker_osc_utils.get_column_definitions(
            vnflcm._attr_map, long_listing=True)

        expected_data = []
        for vnf_instance_obj in vnf_instances:
            expected_data.append(vnflcm_fakes.get_vnflcm_data(
                vnf_instance_obj, columns=columns, list_action=True))

        self.assertCountEqual(_get_columns_vnflcm(action='list'),
                              actual_columns)
        self.assertCountEqual(expected_data, list(data))


class TestInstantiateVnfLcm(TestVnfLcm):

    def setUp(self):
        super(TestInstantiateVnfLcm, self).setUp()
        self.instantiate_vnf_lcm = vnflcm.InstantiateVnfLcm(
            self.app, self.app_args, cmd_name='vnflcm instantiate')

    def test_take_action(self):
        vnf_instance = vnflcm_fakes.vnf_instance_response()
        sample_param_file = ("./tackerclient/osc/v1/vnflcm/samples/"
                             "instantiate_vnf_instance_param_sample.json")

        arglist = [vnf_instance['id'], sample_param_file]
        verifylist = [('vnf_instance', vnf_instance['id']),
                      ('instantiation_request_file', sample_param_file)]

        # command param
        parsed_args = self.check_parser(self.instantiate_vnf_lcm, arglist,
                                        verifylist)

        url = os.path.join(self.url, 'vnflcm/v1/vnf_instances',
                           vnf_instance['id'], 'instantiate')
        self.requests_mock.register_uri(
            'POST', url, headers=self.header, json={})

        sys.stdout = buffer = StringIO()
        with mock.patch.object(proxy_client.ClientBase,
                               '_handle_fault_response') as m:
            self.instantiate_vnf_lcm.take_action(parsed_args)
            # check no fault response is received
            self.assertNotCalled(m)
            self.assertEqual(
                'Instantiate request for VNF Instance ' + vnf_instance['id'] +
                ' has been accepted.', buffer.getvalue().strip())

    def test_take_action_vnf_instance_not_found(self):
        vnf_instance = vnflcm_fakes.vnf_instance_response()
        sample_param_file = ("./tackerclient/osc/v1/vnflcm/samples/"
                             "instantiate_vnf_instance_param_sample.json")
        arglist = [vnf_instance['id'], sample_param_file]
        verifylist = [('vnf_instance', vnf_instance['id']),
                      ('instantiation_request_file', sample_param_file)]

        # command param
        parsed_args = self.check_parser(self.instantiate_vnf_lcm, arglist,
                                        verifylist)

        url = os.path.join(self.url, 'vnflcm/v1/vnf_instances',
                           vnf_instance['id'], 'instantiate')
        self.requests_mock.register_uri(
            'POST', url, headers=self.header, status_code=404, json={})

        self.assertRaises(exceptions.TackerClientException,
                          self.instantiate_vnf_lcm.take_action,
                          parsed_args)

    def test_take_action_param_file_not_exists(self):
        vnf_instance = vnflcm_fakes.vnf_instance_response()
        sample_param_file = "./not_exists.json"
        arglist = [vnf_instance['id'], sample_param_file]
        verifylist = [('vnf_instance', vnf_instance['id']),
                      ('instantiation_request_file', sample_param_file)]

        # command param
        parsed_args = self.check_parser(self.instantiate_vnf_lcm, arglist,
                                        verifylist)

        ex = self.assertRaises(exceptions.InvalidInput,
                               self.instantiate_vnf_lcm.take_action,
                               parsed_args)

        expected_msg = ("Invalid input: "
                        "User does not have read privileges to it")
        self.assertEqual(expected_msg, str(ex))

    @mock.patch("os.open")
    @mock.patch("os.access")
    def test_take_action_invalid_format_param_file(self, mock_open,
                                                   mock_access):
        vnf_instance = vnflcm_fakes.vnf_instance_response()
        sample_param_file = "./invalid_param_file.json"
        arglist = [vnf_instance['id'], sample_param_file]
        verifylist = [('vnf_instance', vnf_instance['id']),
                      ('instantiation_request_file', sample_param_file)]

        mock_open.return_value = "invalid_json_data"
        # command param
        parsed_args = self.check_parser(self.instantiate_vnf_lcm, arglist,
                                        verifylist)

        ex = self.assertRaises(exceptions.InvalidInput,
                               self.instantiate_vnf_lcm.take_action,
                               parsed_args)
        expected_msg = "Failed to load parameter file."
        self.assertIn(expected_msg, str(ex))


@ddt.ddt
class TestHealVnfLcm(TestVnfLcm):

    def setUp(self):
        super(TestHealVnfLcm, self).setUp()
        self.heal_vnf_lcm = vnflcm.HealVnfLcm(
            self.app, self.app_args, cmd_name='vnflcm heal')

    _heal_sample_param_file = ("./tackerclient/osc/v1/vnflcm/samples/"
                               "heal_vnf_instance_param_sample.json")

    @ddt.data((['--cause', 'test-cause', "--vnfc-instance",
                'vnfc-id-1', 'vnfc-id-2'],
               [('cause', 'test-cause'),
                ('vnfc_instance', ['vnfc-id-1', 'vnfc-id-2'])]),
              (['--cause', 'test-cause'],
               [('cause', 'test-cause')]),
              (["--vnfc-instance", 'vnfc-id-1', 'vnfc-id-2'],
               [('vnfc_instance', ['vnfc-id-1', 'vnfc-id-2'])]),
              (["--additional-param-file", _heal_sample_param_file],
               [('additional_param_file', _heal_sample_param_file)]),
              ([], []))
    @ddt.unpack
    def test_take_action(self, arglist, verifylist):
        vnf_instance = vnflcm_fakes.vnf_instance_response()
        arglist.insert(0, vnf_instance['id'])
        verifylist.extend([('vnf_instance', vnf_instance['id'])])

        # command param
        parsed_args = self.check_parser(self.heal_vnf_lcm, arglist,
                                        verifylist)

        url = os.path.join(self.url, 'vnflcm/v1/vnf_instances',
                           vnf_instance['id'], 'heal')
        self.requests_mock.register_uri(
            'POST', url, headers=self.header, json={})

        sys.stdout = buffer = StringIO()
        result_error = self.heal_vnf_lcm.take_action(parsed_args)

        self.assertIsNone(result_error)
        actual_message = buffer.getvalue().strip()

        expected_message = ("Heal request for VNF Instance %s has been "
                            "accepted.") % vnf_instance['id']
        self.assertIn(expected_message, actual_message)

    def test_take_action_vnf_instance_not_found(self):
        vnf_instance = vnflcm_fakes.vnf_instance_response()
        arglist = [vnf_instance['id']]
        verifylist = [('vnf_instance', vnf_instance['id'])]

        # command param
        parsed_args = self.check_parser(self.heal_vnf_lcm, arglist,
                                        verifylist)

        url = os.path.join(self.url, 'vnflcm/v1/vnf_instances',
                           vnf_instance['id'], 'heal')
        self.requests_mock.register_uri(
            'POST', url, headers=self.header, status_code=404, json={})

        self.assertRaises(exceptions.TackerClientException,
                          self.heal_vnf_lcm.take_action,
                          parsed_args)

    def test_take_action_param_file_not_exists(self):
        vnf_instance = vnflcm_fakes.vnf_instance_response()
        sample_param_file = "./not_exists.json"
        arglist = [vnf_instance['id'],
                   '--additional-param-file', sample_param_file]
        verifylist = [('vnf_instance', vnf_instance['id']),
                      ('additional_param_file', sample_param_file)]

        # command param
        parsed_args = self.check_parser(self.heal_vnf_lcm, arglist,
                                        verifylist)

        ex = self.assertRaises(exceptions.InvalidInput,
                               self.heal_vnf_lcm.take_action, parsed_args)

        expected_msg = ("Invalid input: "
                        "User does not have read privileges to it")
        self.assertEqual(expected_msg, str(ex))


@ddt.ddt
class TestTerminateVnfLcm(TestVnfLcm):

    def setUp(self):
        super(TestTerminateVnfLcm, self).setUp()
        self.terminate_vnf_instance = vnflcm.TerminateVnfLcm(
            self.app, self.app_args, cmd_name='vnflcm terminate')

    @ddt.data({'termination_type': 'GRACEFUL', 'delete_vnf': True},
              {'termination_type': 'FORCEFUL', 'delete_vnf': False})
    @ddt.unpack
    def test_take_action(self, termination_type, delete_vnf):
        # argument 'delete_vnf' decides deletion of vnf instance post
        # termination.
        vnf_instance = vnflcm_fakes.vnf_instance_response()
        arglist = ['--termination-type', termination_type, vnf_instance['id']]

        verifylist = [('termination_type', termination_type),
                      ('vnf_instance', vnf_instance['id'])]

        if delete_vnf:
            arglist.extend(['--D'])
            verifylist.extend([('D', True)])

        if termination_type == 'GRACEFUL':
            arglist.extend(['--graceful-termination-timeout', '60'])
            verifylist.append(('graceful_termination_timeout', 60))
        parsed_args = self.check_parser(self.terminate_vnf_instance, arglist,
                                        verifylist)
        url = os.path.join(self.url, 'vnflcm/v1/vnf_instances',
                           vnf_instance['id'], 'terminate')

        with mock.patch.object(proxy_client.ClientBase,
                               '_handle_fault_response') as m:
            self.requests_mock.register_uri('POST', url, json={},
                                            headers=self.header)
            if delete_vnf:
                self.requests_mock.register_uri(
                    'GET', os.path.join(self.url, 'vnflcm/v1/vnf_instances',
                                        vnf_instance['id']),
                    json=vnf_instance, headers=self.header)
                self.requests_mock.register_uri(
                    'DELETE', os.path.join(
                        self.url, 'vnflcm/v1/vnf_instances',
                        vnf_instance['id']), json={}, headers=self.header)

            sys.stdout = buffer = StringIO()
            result = self.terminate_vnf_instance.take_action(parsed_args)
            actual_message = buffer.getvalue().strip()

            expected_message = ("Terminate request for VNF Instance '%s'"
                                " has been accepted.") % vnf_instance['id']
            self.assertIn(expected_message, actual_message)

            if delete_vnf:
                expected_message = ("VNF Instance '%s' is deleted successfully"
                                    % vnf_instance['id'])
                self.assertIn(expected_message, actual_message)

        self.assertIsNone(result)
        self.assertNotCalled(m)

    def test_take_action_terminate_and_delete_wait_failed(self):
        vnf_instance = vnflcm_fakes.vnf_instance_response()
        termination_type = 'GRACEFUL'
        arglist = ['--termination-type', termination_type, '--D',
                   '--graceful-termination-timeout', '5', vnf_instance['id']]

        verifylist = [('termination_type', termination_type), ('D', True),
                      ('graceful_termination_timeout', 5),
                      ('vnf_instance', vnf_instance['id'])]

        parsed_args = self.check_parser(self.terminate_vnf_instance, arglist,
                                        verifylist)
        url = os.path.join(self.url, 'vnflcm/v1/vnf_instances',
                           vnf_instance['id'], 'terminate')

        self.requests_mock.register_uri('POST', url, json={},
                                        headers=self.header)
        # set the instantiateState to "INSTANTIATED", so that the
        # _wait_until_vnf_is_terminated will fail
        vnf_instance['instantiationState'] = 'INSTANTIATED'

        self.requests_mock.register_uri(
            'GET', os.path.join(self.url, 'vnflcm/v1/vnf_instances',
                                vnf_instance['id']),
            json=vnf_instance, headers=self.header)

        sys.stdout = buffer = StringIO()
        with mock.patch.object(self.app.client_manager.tackerclient,
                               'delete_vnf_instance') as mock_delete:
            result = self.assertRaises(
                exceptions.CommandError,
                self.terminate_vnf_instance.take_action, parsed_args)

            actual_message = buffer.getvalue().strip()

            # Terminate vnf instance verification
            expected_message = ("Terminate request for VNF Instance '%s'"
                                " has been accepted.") % vnf_instance['id']
            self.assertIn(expected_message, actual_message)

            # Verify it fails to wait for termination before delete
            expected_message = ("Couldn't verify vnf instance is terminated "
                                "within '%(timeout)s' seconds. Unable to "
                                "delete vnf instance %(id)s"
                                % {'timeout': 15, 'id': vnf_instance['id']})

            self.assertIn(expected_message, str(result))
            self.assertNotCalled(mock_delete)

    def test_terminate_no_options(self):
        self.assertRaises(base.ParserException, self.check_parser,
                          self.terminate_vnf_instance, [], [])

    def test_take_action_vnf_instance_not_found(self):
        vnf_instance = vnflcm_fakes.vnf_instance_response()
        termination_type = 'GRACEFUL'
        arglist = ['--termination-type', termination_type, '--D',
                   '--graceful-termination-timeout', '5', vnf_instance['id']]

        verifylist = [('termination_type', termination_type), ('D', True),
                      ('graceful_termination_timeout', 5),
                      ('vnf_instance', vnf_instance['id'])]

        parsed_args = self.check_parser(self.terminate_vnf_instance, arglist,
                                        verifylist)

        url = os.path.join(self.url, 'vnflcm/v1/vnf_instances',
                           vnf_instance['id'], 'terminate')
        self.requests_mock.register_uri('POST', url, headers=self.header,
                                        status_code=404, json={})

        self.assertRaises(exceptions.TackerClientException,
                          self.terminate_vnf_instance.take_action,
                          parsed_args)


class TestDeleteVnfLcm(TestVnfLcm):

    def setUp(self):
        super(TestDeleteVnfLcm, self).setUp()
        self.delete_vnf_instance = vnflcm.DeleteVnfLcm(
            self.app, self.app_args, cmd_name='vnflcm delete')

        # Vnf Instance to delete
        self.vnf_instances = vnflcm_fakes.create_vnf_instances(count=3)

    def _mock_request_url_for_delete(self, vnf_index):
        url = os.path.join(self.url, 'vnflcm/v1/vnf_instances',
                           self.vnf_instances[vnf_index]['id'])

        json = self.vnf_instances[vnf_index]

        self.requests_mock.register_uri('GET', url, json=json,
                                        headers=self.header)
        self.requests_mock.register_uri('DELETE', url,
                                        headers=self.header, json={})

    def test_delete_one_vnf_instance(self):
        arglist = [self.vnf_instances[0]['id']]
        verifylist = [('vnf_instances',
                       [self.vnf_instances[0]['id']])]

        parsed_args = self.check_parser(self.delete_vnf_instance, arglist,
                                        verifylist)

        self._mock_request_url_for_delete(0)
        sys.stdout = buffer = StringIO()
        result = self.delete_vnf_instance.take_action(parsed_args)
        self.assertIsNone(result)
        self.assertEqual(("Vnf instance '%s' is deleted successfully")
                         % self.vnf_instances[0]['id'],
                         buffer.getvalue().strip())

    def test_delete_multiple_vnf_instance(self):
        arglist = []
        for vnf_pkg in self.vnf_instances:
            arglist.append(vnf_pkg['id'])
        verifylist = [('vnf_instances', arglist)]
        parsed_args = self.check_parser(self.delete_vnf_instance, arglist,
                                        verifylist)
        for i in range(0, 3):
            self._mock_request_url_for_delete(i)
        sys.stdout = buffer = StringIO()
        result = self.delete_vnf_instance.take_action(parsed_args)
        self.assertIsNone(result)
        self.assertEqual('All specified vnf instances are deleted '
                         'successfully', buffer.getvalue().strip())

    def test_delete_multiple_vnf_instance_exception(self):
        arglist = [
            self.vnf_instances[0]['id'],
            'xxxx-yyyy-zzzz',
            self.vnf_instances[1]['id'],
        ]
        verifylist = [('vnf_instances', arglist)]
        parsed_args = self.check_parser(self.delete_vnf_instance,
                                        arglist, verifylist)

        self._mock_request_url_for_delete(0)

        url = os.path.join(self.url, 'vnflcm/v1/vnf_instances',
                           'xxxx-yyyy-zzzz')
        self.requests_mock.register_uri(
            'GET', url, exc=exceptions.ConnectionFailed)

        self._mock_request_url_for_delete(1)
        exception = self.assertRaises(exceptions.CommandError,
                                      self.delete_vnf_instance.take_action,
                                      parsed_args)

        self.assertEqual('Failed to delete 1 of 3 vnf instances.',
                         exception.message)


class TestUpdateVnfLcm(TestVnfLcm):
    def setUp(self):
        super(TestUpdateVnfLcm, self).setUp()
        self.update_vnf_lcm = vnflcm.UpdateVnfLcm(
            self.app, self.app_args, cmd_name='vnflcm modify')

    def test_take_action(self):
        vnf_instance = vnflcm_fakes.vnf_instance_response()
        sample_param_file = ("./tackerclient/osc/v1/vnflcm/samples/"
                             "update_vnf_instance_param_sample.json")

        arglist = [vnf_instance['id'], '--I', sample_param_file]
        verifylist = [('vnf_instance', vnf_instance['id']),
                      ('I', sample_param_file)]

        # command param
        parsed_args = self.check_parser(
            self.update_vnf_lcm, arglist, verifylist)
        url = os.path.join(
            self.url,
            'vnflcm/v1/vnf_instances',
            vnf_instance['id'])

        self.requests_mock.register_uri(
            'PATCH', url, headers=self.header, json={})

        sys.stdout = buffer = StringIO()
        self.update_vnf_lcm.take_action(parsed_args)

        actual_message = buffer.getvalue().strip()

        expected_message = ('Update vnf:' + vnf_instance['id'])

        self.assertEqual(expected_message, actual_message)

    @mock.patch.object(proxy_client.ClientBase, 'deserialize')
    def test_take_action_check_content_type(self, mock_des):
        vnf_instance = vnflcm_fakes.vnf_instance_response()
        sample_param_file = ('./tackerclient/osc/v1/vnflcm/samples/'
                             'update_vnf_instance_param_sample.json')

        arglist = [vnf_instance['id'], '--I', sample_param_file]
        verifylist = [('vnf_instance', vnf_instance['id']),
                      ('I', sample_param_file)]
        mock_des.return_value = {}

        # command param
        parsed_args = self.check_parser(
            self.update_vnf_lcm, arglist, verifylist)
        url = os.path.join(self.url, 'vnflcm/v1/vnf_instances',
                           vnf_instance['id'])

        self.requests_mock.register_uri(
            'PATCH', url, headers=self.header, json={})

        sys.stdout = buffer = StringIO()

        with mock.patch.object(root_client.HTTPClient,
                               'do_request') as mock_req:
            headers = {'Content-Type': 'application/json'}
            mock_req.return_value = (MyResp(202, headers=headers), None)
            self.update_vnf_lcm.take_action(parsed_args)
            # check content_type
            mock_req.assert_called_once_with(
                f'/vnflcm/v1/vnf_instances/{vnf_instance["id"]}', 'PATCH',
                body=mock.ANY, headers=mock.ANY,
                content_type='application/merge-patch+json', accept='json')

        actual_message = buffer.getvalue().strip()

        expected_message = f'Update vnf:{vnf_instance["id"]}'

        self.assertEqual(expected_message, actual_message)

    def test_take_action_param_file_not_exists(self):
        vnf_instance = vnflcm_fakes.vnf_instance_response()
        sample_param_file = "./not_exists.json"
        arglist = [vnf_instance['id'], '--I', sample_param_file]
        verifylist = [('vnf_instance', vnf_instance['id']),
                      ('I', sample_param_file)]

        # command param
        parsed_args = self.check_parser(self.update_vnf_lcm, arglist,
                                        verifylist)

        self.assertRaises(exceptions.InvalidInput,
                          self.update_vnf_lcm.take_action, parsed_args)


@ddt.ddt
class TestScaleVnfLcm(TestVnfLcm):
    def setUp(self):
        super(TestScaleVnfLcm, self).setUp()
        self.scale_vnf_lcm = vnflcm.ScaleVnfLcm(
            self.app, self.app_args, cmd_name='vnflcm scale')

    @ddt.data('SCALE_IN', 'SCALE_OUT')
    def test_take_action(self, scale_type):
        vnf_instance = vnflcm_fakes.vnf_instance_response()
        sample_param_file = ("./tackerclient/osc/v1/vnflcm/samples/"
                             "scale_vnf_instance_param_sample.json")

        arglist = [vnf_instance['id'],
                   '--aspect-id', uuidsentinel.aspect_id,
                   '--number-of-steps', '1',
                   '--type', scale_type,
                   '--additional-param-file', sample_param_file]
        verifylist = [('vnf_instance', vnf_instance['id']),
                      ('aspect_id', uuidsentinel.aspect_id),
                      ('number_of_steps', 1),
                      ('type', scale_type),
                      ('additional_param_file', sample_param_file)]

        # command param
        parsed_args = self.check_parser(self.scale_vnf_lcm, arglist,
                                        verifylist)
        url = os.path.join(
            self.url,
            'vnflcm/v1/vnf_instances',
            vnf_instance['id'],
            'scale')

        self.requests_mock.register_uri(
            'POST', url, headers=self.header, json={})

        sys.stdout = buffer = StringIO()
        self.scale_vnf_lcm.take_action(parsed_args)

        actual_message = buffer.getvalue().strip()

        expected_message = ("Scale request for VNF Instance %s has been "
                            "accepted.") % vnf_instance['id']

        self.assertEqual(expected_message, actual_message)

    @ddt.data('SCALE_IN', 'SCALE_OUT')
    def test_take_action_no_param_file(self, scale_type):
        vnf_instance = vnflcm_fakes.vnf_instance_response()

        arglist = [vnf_instance['id'],
                   '--aspect-id', uuidsentinel.aspect_id,
                   '--number-of-steps', '1',
                   '--type', scale_type]
        verifylist = [('vnf_instance', vnf_instance['id']),
                      ('aspect_id', uuidsentinel.aspect_id),
                      ('number_of_steps', 1),
                      ('type', scale_type)]

        parsed_args = self.check_parser(self.scale_vnf_lcm, arglist,
                                        verifylist)

        url = os.path.join(self.url, 'vnflcm/v1/vnf_instances',
                           vnf_instance['id'], 'scale')

        self.requests_mock.register_uri(
            'POST', url, headers=self.header, json={})

        sys.stdout = buffer = StringIO()
        self.scale_vnf_lcm.take_action(parsed_args)

        actual_message = buffer.getvalue().strip()

        expected_message = ("Scale request for VNF Instance %s has been "
                            "accepted.") % vnf_instance['id']

        self.assertEqual(expected_message, actual_message)

    @ddt.data('SCALE_IN', 'SCALE_OUT')
    def test_take_action_param_file_not_exists(self, scale_type):
        vnf_instance = vnflcm_fakes.vnf_instance_response()
        sample_param_file = "./not_exists.json"
        arglist = [vnf_instance['id'],
                   '--aspect-id', uuidsentinel.aspect_id,
                   '--number-of-steps', '2',
                   '--type', scale_type,
                   '--additional-param-file', sample_param_file]
        verifylist = [('vnf_instance', vnf_instance['id']),
                      ('aspect_id', uuidsentinel.aspect_id),
                      ('number_of_steps', 2),
                      ('type', scale_type),
                      ('additional_param_file', sample_param_file)]

        # command param
        parsed_args = self.check_parser(self.scale_vnf_lcm, arglist,
                                        verifylist)

        ex = self.assertRaises(exceptions.InvalidInput,
                               self.scale_vnf_lcm.take_action, parsed_args)

        expected_msg = ("Invalid input: "
                        "User does not have read privileges to it")
        self.assertEqual(expected_msg, str(ex))

    @ddt.data('SCALE_IN', 'SCALE_OUT')
    def test_take_action_vnf_instance_not_found(self, scale_type):
        vnf_instance = vnflcm_fakes.vnf_instance_response()
        sample_param_file = ("./tackerclient/osc/v1/vnflcm/samples/"
                             "update_vnf_instance_param_sample.json")
        arglist = [vnf_instance['id'],
                   '--aspect-id', uuidsentinel.aspect_id,
                   '--number-of-steps', '3',
                   '--type', scale_type,
                   '--additional-param-file', sample_param_file]
        verifylist = [('vnf_instance', vnf_instance['id']),
                      ('aspect_id', uuidsentinel.aspect_id),
                      ('number_of_steps', 3),
                      ('type', scale_type),
                      ('additional_param_file', sample_param_file)]

        # command param
        parsed_args = self.check_parser(self.scale_vnf_lcm, arglist,
                                        verifylist)

        url = os.path.join(
            self.url,
            'vnflcm/v1/vnf_instances',
            vnf_instance['id'])
        self.requests_mock.register_uri(
            'POST', url, headers=self.header, status_code=404, json={})

        self.assertRaises(exceptions.TackerClientException,
                          self.scale_vnf_lcm.take_action,
                          parsed_args)


class TestChangeExtConnVnfLcm(TestVnfLcm):

    def setUp(self):
        super(TestChangeExtConnVnfLcm, self).setUp()
        self.change_ext_conn_vnf_lcm = vnflcm.ChangeExtConnVnfLcm(
            self.app, self.app_args,
            cmd_name='vnflcm change-ext-conn')

    def test_take_action(self):
        vnf_instance = vnflcm_fakes.vnf_instance_response()
        sample_param_file = ("./tackerclient/osc/v1/vnflcm/samples/"
                             "change_ext_conn_vnf_instance_param_sample.json")

        arglist = [vnf_instance['id'], sample_param_file]
        verifylist = [('vnf_instance', vnf_instance['id']),
                      ('request_file', sample_param_file)]

        # command param
        parsed_args = self.check_parser(self.change_ext_conn_vnf_lcm,
                                        arglist,
                                        verifylist)

        url = os.path.join(self.url, 'vnflcm/v1/vnf_instances',
                           vnf_instance['id'], 'change_ext_conn')
        self.requests_mock.register_uri(
            'POST', url, headers=self.header, json={})

        sys.stdout = buffer = StringIO()
        with mock.patch.object(proxy_client.ClientBase,
                               '_handle_fault_response') as m:
            self.change_ext_conn_vnf_lcm.take_action(parsed_args)
            # check no fault response is received
            self.assertNotCalled(m)
            self.assertEqual(
                ('Change External VNF Connectivity for VNF Instance {0} '
                 'has been accepted.'.format(vnf_instance['id'])),
                buffer.getvalue().strip())

    def test_take_action_vnf_instance_not_found(self):
        vnf_instance = vnflcm_fakes.vnf_instance_response()
        sample_param_file = ("./tackerclient/osc/v1/vnflcm/samples/"
                             "change_ext_conn_vnf_instance_param_sample.json")
        arglist = [vnf_instance['id'], sample_param_file]
        verifylist = [('vnf_instance', vnf_instance['id']),
                      ('request_file', sample_param_file)]

        # command param
        parsed_args = self.check_parser(self.change_ext_conn_vnf_lcm,
                                        arglist,
                                        verifylist)

        url = os.path.join(self.url, 'vnflcm/v1/vnf_instances',
                           vnf_instance['id'], 'change_ext_conn')
        self.requests_mock.register_uri(
            'POST', url, headers=self.header, status_code=404, json={})

        self.assertRaises(exceptions.TackerClientException,
                          self.change_ext_conn_vnf_lcm.take_action,
                          parsed_args)

    def test_take_action_param_file_not_exists(self):
        vnf_instance = vnflcm_fakes.vnf_instance_response()
        sample_param_file = "./not_exists.json"
        arglist = [vnf_instance['id'], sample_param_file]
        verifylist = [('vnf_instance', vnf_instance['id']),
                      ('request_file', sample_param_file)]

        # command param
        parsed_args = self.check_parser(
            self.change_ext_conn_vnf_lcm,
            arglist,
            verifylist)

        ex = self.assertRaises(
            exceptions.InvalidInput,
            self.change_ext_conn_vnf_lcm.take_action,
            parsed_args)

        expected_msg = ("Invalid input: "
                        "User does not have read privileges to it")
        self.assertEqual(expected_msg, str(ex))

    @mock.patch("os.open")
    @mock.patch("os.access")
    def test_take_action_invalid_format_param_file(self, mock_open,
                                                   mock_access):
        vnf_instance = vnflcm_fakes.vnf_instance_response()
        sample_param_file = "./invalid_param_file.json"
        arglist = [vnf_instance['id'], sample_param_file]
        verifylist = [('vnf_instance', vnf_instance['id']),
                      ('request_file', sample_param_file)]

        mock_open.return_value = "invalid_json_data"
        # command param
        parsed_args = self.check_parser(self.change_ext_conn_vnf_lcm,
                                        arglist,
                                        verifylist)

        ex = self.assertRaises(
            exceptions.InvalidInput,
            self.change_ext_conn_vnf_lcm.take_action,
            parsed_args)

        expected_msg = "Failed to load parameter file."
        self.assertIn(expected_msg, str(ex))


class TestChangeVnfPkgVnfLcm(TestVnfLcm):

    def setUp(self):
        super(TestChangeVnfPkgVnfLcm, self).setUp()
        self.change_vnfpkg_vnf_lcm = vnflcm.ChangeVnfPkgVnfLcm(
            self.app, self.app_args,
            cmd_name='vnflcm change-vnfpkg')

    def test_take_action_with_v1_version(self):
        vnf_instance = vnflcm_fakes.vnf_instance_response()
        sample_param_file = ("./tackerclient/osc/v2/vnflcm/samples/"
                             "change_vnfpkg_vnf_instance_param_sample.json")
        arglist = [vnf_instance['id'], sample_param_file]
        verifylist = [('vnf_instance', vnf_instance['id']),
                      ('request_file', sample_param_file)]

        # command param
        parsed_args = self.check_parser(self.change_vnfpkg_vnf_lcm,
                                        arglist,
                                        verifylist)

        url = os.path.join(self.url, 'vnflcm/v1/vnf_instances',
                           vnf_instance['id'], 'change_vnfpkg')
        self.requests_mock.register_uri(
            'POST', url, headers=self.header, status_code=400, json={})

        ex = self.assertRaises(exceptions.UnsupportedCommandVersion,
                               self.change_vnfpkg_vnf_lcm.take_action,
                               parsed_args)

        expected_msg = "This command is not supported in version 1"
        self.assertEqual(expected_msg, str(ex))


class TestVnfLcmV1(base.FixturedTestCase):
    client_fixture_class = client.ClientFixture
    api_version = '1'

    def setUp(self):
        super(TestVnfLcmV1, self).setUp()

    def test_client_v2(self):
        self.assertEqual(self.cs.vnf_lcm_client.headers,
                         {'Version': '1.3.0'})
        self.assertEqual(self.cs.vnf_lcm_client.vnf_instances_path,
                         '/vnflcm/v1/vnf_instances')
        # check of other paths is omitted.



# Copyright 2019 NTT DATA
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import fixtures
from keystoneauth1 import fixture
from keystoneauth1 import loading
from keystoneauth1 import session

from tackerclient.v1_0 import client as proxy_client

IDENTITY_URL = 'http://identityserver:5000/v3'
TACKER_URL = 'http://nfv-orchestration'


class ClientFixture(fixtures.Fixture):

    def __init__(self, requests_mock, identity_url=IDENTITY_URL,
                 api_version='1'):
        super(ClientFixture, self).__init__()
        self.identity_url = identity_url
        self.client = None
        self.token = fixture.V2Token()
        self.token.set_scope()
        self.requests_mock = requests_mock
        self.discovery = fixture.V2Discovery(href=self.identity_url)
        s = self.token.add_service('nfv-orchestration')
        s.add_endpoint(TACKER_URL)
        self.api_version = api_version

    def setUp(self):
        super(ClientFixture, self).setUp()
        auth_url = '%s/tokens' % self.identity_url
        headers = {'X-Content-Type': 'application/json'}
        self.requests_mock.post(auth_url, json=self.token, headers=headers)
        self.requests_mock.get(self.identity_url, json=self.discovery,
                               headers=headers)
        self.client = self.new_client()

    def new_client(self):
        self.session = session.Session()
        loader = loading.get_plugin_loader('password')
        self.session.auth = loader.load_from_options(
            auth_url=self.identity_url, username='xx', password='xx')

        return proxy_client.Client(service_type='nfv-orchestration',
                                   interface='public',
                                   endpoint_type='public',
                                   region_name='RegionOne',
                                   auth_url=self.identity_url,
                                   token=self.token.token_id,
                                   endpoint_url=TACKER_URL,
                                   api_version=self.api_version)



# Copyright (C) 2013 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import fixtures
from keystoneclient import session
import requests
import testtools

from tackerclient import client
from tackerclient.common import clientmanager
from tackerclient.common import exceptions
from tackerclient import shell as openstack_shell


AUTH_TOKEN = 'test_token'
END_URL = 'test_url'
METHOD = 'GET'
URL = 'http://test.test:1234/v1.0/'
CA_CERT = '/tmp/test/path'
DEFAULT_API_VERSION = '1.0'


class TestSSL(testtools.TestCase):
    def setUp(self):
        super(TestSSL, self).setUp()

        self.useFixture(fixtures.EnvironmentVariable('OS_TOKEN', AUTH_TOKEN))
        self.useFixture(fixtures.EnvironmentVariable('OS_URL', END_URL))
        self.addCleanup(mock.patch.stopall)

    def _test_verify_client_manager(self, cacert):
        with mock.patch.object(session, 'Session'), \
                mock.patch.object(clientmanager, 'ClientManager') as mock_cmgr:

            mock_cmgr.return_value = 0
            shell = openstack_shell.TackerShell(DEFAULT_API_VERSION)
            shell.options = mock.Mock()
            auth_session = shell._get_keystone_session()

            shell.run(cacert)

        mock_cmgr.assert_called_with(
            api_version={'nfv-orchestration': '1.0'},
            auth=auth_session.auth, auth_strategy='keystone',
            auth_url='', ca_cert=CA_CERT, endpoint_type='publicURL',
            insecure=False, log_credentials=True, password='',
            raise_errors=False, region_name='', retries=0,
            service_type='nfv-orchestration', session=auth_session,
            tenant_id='', tenant_name='', timeout=None,
            token='test_token', url='test_url', user_id='', username='')

    def test_ca_cert_passed(self):
        cacert = ['--os-cacert', CA_CERT]
        self._test_verify_client_manager(cacert)

    def test_ca_cert_passed_as_env_var(self):
        self.useFixture(fixtures.EnvironmentVariable('OS_CACERT', CA_CERT))
        self._test_verify_client_manager([])

    @mock.patch.object(client.HTTPClient, 'request')
    def test_proper_exception_is_raised_when_cert_validation_fails(self,
                                                                   mock_req):
        http = client.HTTPClient(token=AUTH_TOKEN, endpoint_url=END_URL)
        mock_req.side_effect = requests.exceptions.SSLError()
        self.assertRaises(
            exceptions.SslCertificateValidationError,
            http._cs_request,
            URL, METHOD
        )



# Copyright 2012 OpenStack Foundation.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from sys import stderr

from cliff import command


class OpenStackCommand(command.Command):
    """Base class for OpenStack commands."""

    api = None

    def run(self, parsed_args):
        stderr.write("Deprecated: tacker command line is deprecated, "
                     "will be deleted after Rocky is released.\n")
        if not self.api:
            return
        else:
            return super(OpenStackCommand, self).run(parsed_args)

    def get_data(self, parsed_args):
        pass

    def take_action(self, parsed_args):
        return self.get_data(parsed_args)



# Copyright 2012 OpenStack Foundation.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""
Command-line interface to the Tacker APIs
"""

import argparse
import getpass
import inspect
import itertools
import logging
import os
import sys
from urllib import parse as urlparse

from cliff import app
from cliff import commandmanager

from keystoneclient.auth.identity import v2 as v2_auth
from keystoneclient.auth.identity import v3 as v3_auth
from keystoneclient import discover
from keystoneclient import exceptions as ks_exc
from keystoneclient import session
from oslo_utils import encodeutils

from tackerclient.common import clientmanager
from tackerclient.common import command as openstack_command
from tackerclient.common import exceptions as exc
from tackerclient.common import extension as client_extension
from tackerclient.common import utils
from tackerclient.i18n import _
from tackerclient.tacker.v1_0.nfvo import vim
from tackerclient.version import __version__


VERSION = '1.0'
TACKER_API_VERSION = '1.0'


def run_command(cmd, cmd_parser, sub_argv):
    _argv = sub_argv
    index = -1
    values_specs = []
    if '--' in sub_argv:
        index = sub_argv.index('--')
        _argv = sub_argv[:index]
        values_specs = sub_argv[index:]
    known_args, _values_specs = cmd_parser.parse_known_args(_argv)
    cmd.values_specs = (index == -1 and _values_specs or values_specs)
    return cmd.run(known_args)


def env(*_vars, **kwargs):
    """Search for the first defined of possibly many env vars.

    Returns the first environment variable defined in vars, or
    returns the default defined in kwargs.

    """
    for v in _vars:
        value = os.environ.get(v, None)
        if value:
            return value
    return kwargs.get('default', '')


def check_non_negative_int(value):
    try:
        value = int(value)
    except ValueError:
        raise argparse.ArgumentTypeError(_("invalid int value: %r") % value)
    if value < 0:
        raise argparse.ArgumentTypeError(_("input value %d is negative") %
                                         value)
    return value


class BashCompletionCommand(openstack_command.OpenStackCommand):
    """Prints all of the commands and options for bash-completion."""
    resource = "bash_completion"


COMMAND_V1 = {
    'bash-completion': BashCompletionCommand,

    # MANO lingo
    'vim-register': vim.CreateVIM,
    'vim-update': vim.UpdateVIM,
    'vim-delete': vim.DeleteVIM,
    'vim-list': vim.ListVIM,
    'vim-show': vim.ShowVIM
}

COMMANDS = {'1.0': COMMAND_V1}


class HelpAction(argparse.Action):
    """Provides a custom action for the -h and --help options.

    The commands are determined by checking the CommandManager
    instance, passed in as the "default" value for the action.

    :returns: a list of the commands
    """
    def __call__(self, parser, namespace, values, option_string=None):
        outputs = []
        max_len = 0
        app = self.default
        parser.print_help(app.stdout)
        app.stdout.write(_('\nCommands for API v%s:\n') % app.api_version)
        command_manager = app.command_manager
        for name, ep in sorted(command_manager):
            factory = ep.load()
            cmd = factory(self, None)
            one_liner = cmd.get_description().split('\n')[0]
            outputs.append((name, one_liner))
            max_len = max(len(name), max_len)
        for (name, one_liner) in outputs:
            app.stdout.write('  %s  %s\n' % (name.ljust(max_len), one_liner))
        sys.exit(0)


class TackerShell(app.App):

    # verbose logging levels
    WARNING_LEVEL = 0
    INFO_LEVEL = 1
    DEBUG_LEVEL = 2
    CONSOLE_MESSAGE_FORMAT = '%(message)s'
    DEBUG_MESSAGE_FORMAT = '%(levelname)s: %(name)s %(message)s'
    log = logging.getLogger(__name__)

    def __init__(self, apiversion):
        super(TackerShell, self).__init__(
            description=__doc__.strip(),
            version=VERSION,
            command_manager=commandmanager.CommandManager('tacker.cli'), )
        self.commands = COMMANDS
        for k, v in self.commands[apiversion].items():
            self.command_manager.add_command(k, v)

        self._register_extensions(VERSION)

        # Pop the 'complete' to correct the outputs of 'tacker help'.
        self.command_manager.commands.pop('complete')

        # This is instantiated in initialize_app() only when using
        # password flow auth
        self.auth_client = None
        self.api_version = apiversion

    def build_option_parser(self, description, version):
        """Return an argparse option parser for this application.

        Subclasses may override this method to extend
        the parser with more global options.

        :param description: full description of the application
        :paramtype description: str
        :param version: version number for the application
        :paramtype version: str
        """
        parser = argparse.ArgumentParser(
            description=description,
            add_help=False, )
        parser.add_argument(
            '--version',
            action='version',
            version=__version__, )
        parser.add_argument(
            '-v', '--verbose', '--debug',
            action='count',
            dest='verbose_level',
            default=self.DEFAULT_VERBOSE_LEVEL,
            help=_('Increase verbosity of output and show tracebacks on'
                   ' errors. You can repeat this option.'))
        parser.add_argument(
            '-q', '--quiet',
            action='store_const',
            dest='verbose_level',
            const=0,
            help=_('Suppress output except warnings and errors.'))
        parser.add_argument(
            '-h', '--help',
            action=HelpAction,
            nargs=0,
            default=self,  # tricky
            help=_("Show this help message and exit."))
        parser.add_argument(
            '-r', '--retries',
            metavar="NUM",
            type=check_non_negative_int,
            default=0,
            help=_("How many times the request to the Tacker server should "
                   "be retried if it fails."))
        # FIXME(bklei): this method should come from python-keystoneclient
        self._append_global_identity_args(parser)

        return parser

    def _append_global_identity_args(self, parser):
        # FIXME(bklei): these are global identity (Keystone) arguments which
        # should be consistent and shared by all service clients. Therefore,
        # they should be provided by python-keystoneclient. We will need to
        # refactor this code once this functionality is available in
        # python-keystoneclient.
        #
        # Note: At that time we'll need to decide if we can just abandon
        #       the deprecated args (--service-type and --endpoint-type).

        parser.add_argument(
            '--os-service-type', metavar='<os-service-type>',
            default=env('OS_TACKER_SERVICE_TYPE',
                        default='nfv-orchestration'),
            help=_('Defaults to env[OS_TACKER_SERVICE_TYPE] or \
                    nfv-orchestration.'))

        parser.add_argument(
            '--os-endpoint-type', metavar='<os-endpoint-type>',
            default=env('OS_ENDPOINT_TYPE', default='publicURL'),
            help=_('Defaults to env[OS_ENDPOINT_TYPE] or publicURL.'))

        # FIXME(bklei): --service-type is deprecated but kept in for
        # backward compatibility.
        parser.add_argument(
            '--service-type', metavar='<service-type>',
            default=env('OS_TACKER_SERVICE_TYPE',
                        default='nfv-orchestration'),
            help=_('DEPRECATED! Use --os-service-type.'))

        # FIXME(bklei): --endpoint-type is deprecated but kept in for
        # backward compatibility.
        parser.add_argument(
            '--endpoint-type', metavar='<endpoint-type>',
            default=env('OS_ENDPOINT_TYPE', default='publicURL'),
            help=_('DEPRECATED! Use --os-endpoint-type.'))

        parser.add_argument(
            '--os-auth-strategy', metavar='<auth-strategy>',
            default=env('OS_AUTH_STRATEGY', default='keystone'),
            help=_('DEPRECATED! Only keystone is supported.'))

        parser.add_argument(
            '--os_auth_strategy',
            help=argparse.SUPPRESS)

        parser.add_argument(
            '--os-auth-url', metavar='<auth-url>',
            default=env('OS_AUTH_URL'),
            help=_('Authentication URL, defaults to env[OS_AUTH_URL].'))
        parser.add_argument(
            '--os_auth_url',
            help=argparse.SUPPRESS)

        project_name_group = parser.add_mutually_exclusive_group()
        project_name_group.add_argument(
            '--os-tenant-name', metavar='<auth-tenant-name>',
            default=env('OS_TENANT_NAME'),
            help=_('Authentication tenant name, defaults to '
                   'env[OS_TENANT_NAME].'))
        project_name_group.add_argument(
            '--os-project-name',
            metavar='<auth-project-name>',
            default=utils.env('OS_PROJECT_NAME'),
            help=_('Another way to specify tenant name. '
                   'This option is mutually exclusive with '
                   ' --os-tenant-name. '
                   'Defaults to env[OS_PROJECT_NAME].'))

        parser.add_argument(
            '--os_tenant_name',
            help=argparse.SUPPRESS)

        project_id_group = parser.add_mutually_exclusive_group()
        project_id_group.add_argument(
            '--os-tenant-id', metavar='<auth-tenant-id>',
            default=env('OS_TENANT_ID'),
            help=_('Authentication tenant ID, defaults to '
                   'env[OS_TENANT_ID].'))
        project_id_group.add_argument(
            '--os-project-id',
            metavar='<auth-project-id>',
            default=utils.env('OS_PROJECT_ID'),
            help=_('Another way to specify tenant ID. '
                   'This option is mutually exclusive with '
                   ' --os-tenant-id. '
                   'Defaults to env[OS_PROJECT_ID].'))

        parser.add_argument(
            '--os-username', metavar='<auth-username>',
            default=utils.env('OS_USERNAME'),
            help=_('Authentication username, defaults to env[OS_USERNAME].'))
        parser.add_argument(
            '--os_username',
            help=argparse.SUPPRESS)

        parser.add_argument(
            '--os-user-id', metavar='<auth-user-id>',
            default=env('OS_USER_ID'),
            help=_('Authentication user ID (Env: OS_USER_ID)'))

        parser.add_argument(
            '--os_user_id',
            help=argparse.SUPPRESS)

        parser.add_argument(
            '--os-user-domain-id',
            metavar='<auth-user-domain-id>',
            default=utils.env('OS_USER_DOMAIN_ID'),
            help=_('OpenStack user domain ID. '
                   'Defaults to env[OS_USER_DOMAIN_ID].'))

        parser.add_argument(
            '--os_user_domain_id',
            help=argparse.SUPPRESS)

        parser.add_argument(
            '--os-user-domain-name',
            metavar='<auth-user-domain-name>',
            default=utils.env('OS_USER_DOMAIN_NAME'),
            help=_('OpenStack user domain name. '
                   'Defaults to env[OS_USER_DOMAIN_NAME].'))

        parser.add_argument(
            '--os_user_domain_name',
            help=argparse.SUPPRESS)

        parser.add_argument(
            '--os_project_id',
            help=argparse.SUPPRESS)

        parser.add_argument(
            '--os_project_name',
            help=argparse.SUPPRESS)

        parser.add_argument(
            '--os-project-domain-id',
            metavar='<auth-project-domain-id>',
            default=utils.env('OS_PROJECT_DOMAIN_ID'),
            help=_('Defaults to env[OS_PROJECT_DOMAIN_ID].'))

        parser.add_argument(
            '--os-project-domain-name',
            metavar='<auth-project-domain-name>',
            default=utils.env('OS_PROJECT_DOMAIN_NAME'),
            help=_('Defaults to env[OS_PROJECT_DOMAIN_NAME].'))

        parser.add_argument(
            '--os-cert',
            metavar='<certificate>',
            default=utils.env('OS_CERT'),
            help=_("Path of certificate file to use in SSL "
                   "connection. This file can optionally be "
                   "prepended with the private key. Defaults "
                   "to env[OS_CERT]."))

        parser.add_argument(
            '--os-cacert',
            metavar='<ca-certificate>',
            default=env('OS_CACERT', default=None),
            help=_("Specify a CA bundle file to use in "
                   "verifying a TLS (https) server certificate. "
                   "Defaults to env[OS_CACERT]."))

        parser.add_argument(
            '--os-key',
            metavar='<key>',
            default=utils.env('OS_KEY'),
            help=_("Path of client key to use in SSL "
                   "connection. This option is not necessary "
                   "if your key is prepended to your certificate "
                   "file. Defaults to env[OS_KEY]."))

        parser.add_argument(
            '--os-password', metavar='<auth-password>',
            default=utils.env('OS_PASSWORD'),
            help=_('Authentication password, defaults to env[OS_PASSWORD].'))
        parser.add_argument(
            '--os_password',
            help=argparse.SUPPRESS)

        parser.add_argument(
            '--os-region-name', metavar='<auth-region-name>',
            default=env('OS_REGION_NAME'),
            help=_('Authentication region name, defaults to '
                   'env[OS_REGION_NAME].'))
        parser.add_argument(
            '--os_region_name',
            help=argparse.SUPPRESS)

        parser.add_argument(
            '--os-token', metavar='<token>',
            default=env('OS_TOKEN'),
            help=_('Authentication token, defaults to env[OS_TOKEN].'))
        parser.add_argument(
            '--os_token',
            help=argparse.SUPPRESS)

        parser.add_argument(
            '--http-timeout', metavar='<seconds>',
            default=env('OS_NETWORK_TIMEOUT', default=None), type=float,
            help=_('Timeout in seconds to wait for an HTTP response. Defaults '
                   'to env[OS_NETWORK_TIMEOUT] or None if not specified.'))

        parser.add_argument(
            '--os-url', metavar='<url>',
            default=env('OS_URL'),
            help=_('Defaults to env[OS_URL].'))
        parser.add_argument(
            '--os_url',
            help=argparse.SUPPRESS)

        parser.add_argument(
            '--insecure',
            action='store_true',
            default=env('TACKERCLIENT_INSECURE', default=False),
            help=_("Explicitly allow tackerclient to perform \"insecure\" "
                   "SSL (https) requests. The server's certificate will "
                   "not be verified against any certificate authorities. "
                   "This option should be used with caution."))

    def _bash_completion(self):
        """Prints all of the commands and options for bash-completion."""
        commands = set()
        options = set()
        for option, _action in self.parser._option_string_actions.items():
            options.add(option)
        for command_name, command in self.command_manager:
            commands.add(command_name)
            cmd_factory = command.load()
            cmd = cmd_factory(self, None)
            cmd_parser = cmd.get_parser('')
            for option, _action in cmd_parser._option_string_actions.items():
                options.add(option)
        print(' '.join(commands | options))

    def _register_extensions(self, version):
        for name, module in itertools.chain(
                client_extension._discover_via_entry_points()):
            self._extend_shell_commands(module, version)

    def _extend_shell_commands(self, module, version):
        classes = inspect.getmembers(module, inspect.isclass)
        for cls_name, cls in classes:
            if (issubclass(cls, client_extension.TackerClientExtension) and
                    hasattr(cls, 'shell_command')):
                cmd = cls.shell_command
                if hasattr(cls, 'versions'):
                    if version not in cls.versions:
                        continue
                try:
                    self.command_manager.add_command(cmd, cls)
                    self.commands[version][cmd] = cls
                except TypeError:
                    pass

    def run(self, argv):
        """Equivalent to the main program for the application.

        :param argv: input arguments and options
        :paramtype argv: list of str
        """
        try:
            index = 0
            command_pos = -1
            help_pos = -1
            help_command_pos = -1
            for arg in argv:
                if arg == 'bash-completion' and help_command_pos == -1:
                    self._bash_completion()
                    return 0
                if arg in self.commands[self.api_version]:
                    if command_pos == -1:
                        command_pos = index
                elif arg in ('-h', '--help'):
                    if help_pos == -1:
                        help_pos = index
                elif arg == 'help':
                    if help_command_pos == -1:
                        help_command_pos = index
                index = index + 1
            if command_pos > -1 and help_pos > command_pos:
                argv = ['help', argv[command_pos]]
            if help_command_pos > -1 and command_pos == -1:
                argv[help_command_pos] = '--help'
            self.options, remainder = self.parser.parse_known_args(argv)
            self.configure_logging()
            self.interactive_mode = not remainder
            self.initialize_app(remainder)
        except Exception as err:
            if self.options.verbose_level >= self.DEBUG_LEVEL:
                self.log.exception(err)
                raise
            else:
                self.log.error(err)
            return 1
        if self.interactive_mode:
            _argv = [sys.argv[0]]
            sys.argv = _argv
            return self.interact()
        return self.run_subcommand(remainder)

    def run_subcommand(self, argv):
        subcommand = self.command_manager.find_command(argv)
        cmd_factory, cmd_name, sub_argv = subcommand
        cmd = cmd_factory(self, self.options)
        try:
            self.prepare_to_run_command(cmd)
            full_name = (cmd_name
                         if self.interactive_mode
                         else ' '.join([self.NAME, cmd_name])
                         )
            cmd_parser = cmd.get_parser(full_name)
            return run_command(cmd, cmd_parser, sub_argv)
        except Exception as e:
            if self.options.verbose_level >= self.DEBUG_LEVEL:
                self.log.exception("%s", e)
                raise
            self.log.error("%s", e)
        return 1

    def authenticate_user(self):
        """Authentication validation.

        Make sure the user has provided all of the authentication
        info we need.
        """
        if self.options.os_auth_strategy == 'keystone':
            if self.options.os_token or self.options.os_url:
                # Token flow auth takes priority
                if not self.options.os_token:
                    raise exc.CommandError(
                        _("You must provide a token via"
                          " either --os-token or env[OS_TOKEN]"
                          " when providing a service URL"))

                if not self.options.os_url:
                    raise exc.CommandError(
                        _("You must provide a service URL via"
                          " either --os-url or env[OS_URL]"
                          " when providing a token"))

            else:
                # Validate password flow auth
                project_info = (self.options.os_tenant_name or
                                self.options.os_tenant_id or
                                (self.options.os_project_name and
                                    (self.options.os_project_domain_name or
                                     self.options.os_project_domain_id)) or
                                self.options.os_project_id)

                if (not self.options.os_username and
                        not self.options.os_user_id):
                    raise exc.CommandError(
                        _("You must provide a username or user ID via"
                          "  --os-username, env[OS_USERNAME] or"
                          "  --os-user-id, env[OS_USER_ID]"))

                if not self.options.os_password:
                    # No password, If we've got a tty, try prompting for it
                    if hasattr(sys.stdin, 'isatty') and sys.stdin.isatty():
                        # Check for Ctl-D
                        try:
                            self.options.os_password = getpass.getpass(
                                'OS Password: ')
                        except EOFError:
                            pass
                    # No password because we didn't have a tty or the
                    # user Ctl-D when prompted.
                    if not self.options.os_password:
                        raise exc.CommandError(
                            _("You must provide a password via"
                              " either --os-password or env[OS_PASSWORD]"))

                if (not project_info):
                    # tenent is deprecated in Keystone v3. Use the latest
                    # terminology instead.
                    raise exc.CommandError(
                        _("You must provide a project_id or project_name ("
                          "with project_domain_name or project_domain_id) "
                          "via "
                          "  --os-project-id (env[OS_PROJECT_ID])"
                          "  --os-project-name (env[OS_PROJECT_NAME]),"
                          "  --os-project-domain-id "
                          "(env[OS_PROJECT_DOMAIN_ID])"
                          "  --os-project-domain-name "
                          "(env[OS_PROJECT_DOMAIN_NAME])"))

                if not self.options.os_auth_url:
                    raise exc.CommandError(
                        _("You must provide an auth url via"
                          " either --os-auth-url or via env[OS_AUTH_URL]"))
            auth_session = self._get_keystone_session()
            auth = auth_session.auth
        else:   # not keystone
            if not self.options.os_url:
                raise exc.CommandError(
                    _("You must provide a service URL via"
                      " either --os-url or env[OS_URL]"))
            auth_session = None
            auth = None

        self.client_manager = clientmanager.ClientManager(
            token=self.options.os_token,
            url=self.options.os_url,
            auth_url=self.options.os_auth_url,
            tenant_name=self.options.os_tenant_name,
            tenant_id=self.options.os_tenant_id,
            username=self.options.os_username,
            user_id=self.options.os_user_id,
            password=self.options.os_password,
            region_name=self.options.os_region_name,
            api_version=self.api_version,
            auth_strategy=self.options.os_auth_strategy,
            # FIXME (bklei) honor deprecated service_type and
            # endpoint type until they are removed
            service_type=self.options.os_service_type or
            self.options.service_type,
            endpoint_type=self.options.os_endpoint_type or self.endpoint_type,
            insecure=self.options.insecure,
            ca_cert=self.options.os_cacert,
            timeout=self.options.http_timeout,
            retries=self.options.retries,
            raise_errors=False,
            session=auth_session,
            auth=auth,
            log_credentials=True)
        return

    def initialize_app(self, argv):
        """Global app init bits:

        * set up API versions
        * validate authentication info
        """

        super(TackerShell, self).initialize_app(argv)

        self.api_version = {'nfv-orchestration':
                            self.api_version}

        # If the user is not asking for help, make sure they
        # have given us auth.
        cmd_name = None
        if argv:
            cmd_info = self.command_manager.find_command(argv)
            cmd_factory, cmd_name, sub_argv = cmd_info
        if self.interactive_mode or cmd_name != 'help':
            self.authenticate_user()

    def configure_logging(self):
        """Create logging handlers for any log output."""
        root_logger = logging.getLogger('')

        # Set up logging to a file
        root_logger.setLevel(logging.DEBUG)

        # Send higher-level messages to the console via stderr
        console = logging.StreamHandler(self.stderr)
        console_level = {self.WARNING_LEVEL: logging.WARNING,
                         self.INFO_LEVEL: logging.INFO,
                         self.DEBUG_LEVEL: logging.DEBUG,
                         }.get(self.options.verbose_level, logging.DEBUG)
        # The default log level is INFO, in this situation, set the
        # log level of the console to WARNING, to avoid displaying
        # useless messages. This equals using "--quiet"
        if console_level == logging.INFO:
            console.setLevel(logging.WARNING)
        else:
            console.setLevel(console_level)
        if logging.DEBUG == console_level:
            formatter = logging.Formatter(self.DEBUG_MESSAGE_FORMAT)
        else:
            formatter = logging.Formatter(self.CONSOLE_MESSAGE_FORMAT)
        logging.getLogger('iso8601.iso8601').setLevel(logging.WARNING)
        logging.getLogger('urllib3.connectionpool').setLevel(logging.WARNING)
        console.setFormatter(formatter)
        root_logger.addHandler(console)
        return

    def get_v2_auth(self, v2_auth_url):
        return v2_auth.Password(
            v2_auth_url,
            username=self.options.os_username,
            password=self.options.os_password,
            tenant_id=self.options.os_tenant_id,
            tenant_name=self.options.os_tenant_name)

    def get_v3_auth(self, v3_auth_url):
        project_id = self.options.os_project_id or self.options.os_tenant_id
        project_name = (self.options.os_project_name or
                        self.options.os_tenant_name)

        return v3_auth.Password(
            v3_auth_url,
            username=self.options.os_username,
            password=self.options.os_password,
            user_id=self.options.os_user_id,
            user_domain_name=self.options.os_user_domain_name,
            user_domain_id=self.options.os_user_domain_id,
            project_id=project_id,
            project_name=project_name,
            project_domain_name=self.options.os_project_domain_name,
            project_domain_id=self.options.os_project_domain_id
        )

    def _discover_auth_versions(self, session, auth_url):
        # discover the API versions the server is supporting base on the
        # given URL
        try:
            ks_discover = discover.Discover(session=session, auth_url=auth_url)
            return (ks_discover.url_for('2.0'), ks_discover.url_for('3.0'))
        except ks_exc.ClientException:
            # Identity service may not support discover API version.
            # Lets try to figure out the API version from the original URL.
            url_parts = urlparse.urlparse(auth_url)
            (scheme, netloc, path, params, query, fragment) = url_parts
            path = path.lower()
            if path.startswith('/v3'):
                return (None, auth_url)
            elif path.startswith('/v2'):
                return (auth_url, None)
            else:
                # not enough information to determine the auth version
                msg = _('Unable to determine the Keystone version '
                        'to authenticate with using the given '
                        'auth_url. Identity service may not support API '
                        'version discovery. Please provide a versioned '
                        'auth_url instead.')
                raise exc.CommandError(msg)

    def _get_keystone_session(self):
        # first create a Keystone session
        cacert = self.options.os_cacert or None
        cert = self.options.os_cert or None
        key = self.options.os_key or None
        insecure = self.options.insecure or False
        ks_session = session.Session.construct(dict(cacert=cacert,
                                                    cert=cert,
                                                    key=key,
                                                    insecure=insecure))
        # discover the supported keystone versions using the given url
        (v2_auth_url, v3_auth_url) = self._discover_auth_versions(
            session=ks_session,
            auth_url=self.options.os_auth_url)

        # Determine which authentication plugin to use. First inspect the
        # auth_url to see the supported version. If both v3 and v2 are
        # supported, then use the highest version if possible.
        user_domain_name = self.options.os_user_domain_name or None
        user_domain_id = self.options.os_user_domain_id or None
        project_domain_name = self.options.os_project_domain_name or None
        project_domain_id = self.options.os_project_domain_id or None
        domain_info = (user_domain_name or user_domain_id or
                       project_domain_name or project_domain_id)

        if (v2_auth_url and not domain_info) or not v3_auth_url:
            ks_session.auth = self.get_v2_auth(v2_auth_url)
        else:
            ks_session.auth = self.get_v3_auth(v3_auth_url)

        return ks_session


def main(argv=sys.argv[1:]):
    try:
        return TackerShell(TACKER_API_VERSION).run(
            list(map(encodeutils.safe_decode, argv)))
    except KeyboardInterrupt:
        print("... terminating tacker client", file=sys.stderr)
        return 130
    except exc.TackerClientException:
        return 1
    except Exception as e:
        print(e)
        return 1


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from io import StringIO
import os
import sys

import ddt
from oslo_utils.fixture import uuidsentinel
from unittest import mock

from tackerclient.common import exceptions
from tackerclient.osc import utils as tacker_osc_utils
from tackerclient.osc.v1.vnflcm import vnflcm_op_occs
from tackerclient.tests.unit.osc import base
from tackerclient.tests.unit.osc.v1.fixture_data import client
from tackerclient.tests.unit.osc.v1 import vnflcm_op_occs_fakes


def _get_columns_vnflcm_op_occs(action='show', parameter=None):

    if action == 'fail':
        return ['ID', 'Operation State', 'State Entered Time',
                'Start Time', 'VNF Instance ID', 'Operation',
                'Is Automatic Invocation', 'Is Cancel Pending',
                'Error', 'Links']
    elif action == 'list':
        if parameter is not None:
            return ['ID', 'Operation']
        else:
            return ['ID', 'Operation State', 'VNF Instance ID',
                    'Operation']
    else:
        return ['ID', 'Operation State', 'State Entered Time',
                'Start Time', 'VNF Instance ID', 'Grant ID',
                'Operation', 'Is Automatic Invocation',
                'Operation Parameters', 'Is Cancel Pending',
                'Cancel Mode', 'Error', 'Resource Changes',
                'Changed Info', 'Changed External Connectivity', 'Links']


class TestVnfLcm(base.FixturedTestCase):
    client_fixture_class = client.ClientFixture

    def setUp(self):
        super(TestVnfLcm, self).setUp()
        self.url = client.TACKER_URL
        self.header = {'content-type': 'application/json'}
        self.app = mock.Mock()
        self.app_args = mock.Mock()
        self.client_manager = self.cs
        self.app.client_manager.tackerclient = self.client_manager


@ddt.ddt
class TestCancelVnfLcmOp(TestVnfLcm):

    def setUp(self):
        super(TestCancelVnfLcmOp, self).setUp()
        self.cancel_vnf_lcm = vnflcm_op_occs.CancelVnfLcmOp(
            self.app, self.app_args, cmd_name='vnflcm op cancel')

    @ddt.data('GRACEFUL', 'FORCEFUL')
    def test_take_action(self, cancel_mode):
        """take_action normal system test"""

        arglist = ['--cancel-mode', cancel_mode,
                   uuidsentinel.vnf_lcm_op_occ_id]
        verifylist = [('cancel_mode', cancel_mode),
                      ('vnf_lcm_op_occ_id', uuidsentinel.vnf_lcm_op_occ_id)]

        parsed_args = self.check_parser(
            self.cancel_vnf_lcm, arglist, verifylist)
        url = os.path.join(
            self.url,
            'vnflcm/v1/vnf_lcm_op_occs',
            uuidsentinel.vnf_lcm_op_occ_id,
            'cancel')
        self.requests_mock.register_uri(
            'POST', url, headers=self.header, json={})

        sys.stdout = buffer = StringIO()
        self.cancel_vnf_lcm.take_action(parsed_args)

        actual_message = buffer.getvalue().strip()

        expected_message = (
            'Cancel request for LCM operation ' +
            uuidsentinel.vnf_lcm_op_occ_id +
            ' has been accepted')

        self.assertEqual(expected_message, actual_message)

    def test_terminate_no_options(self):
        self.assertRaises(base.ParserException, self.check_parser,
                          self.cancel_vnf_lcm, [], [])

    def test_take_action_vnf_lcm_op_occ_id_not_found(self):
        """take_action abnomaly system test"""

        arglist = [uuidsentinel.vnf_lcm_op_occ_id]
        verifylist = [('vnf_lcm_op_occ_id', uuidsentinel.vnf_lcm_op_occ_id)]

        parsed_args = self.check_parser(
            self.cancel_vnf_lcm, arglist, verifylist)
        url = os.path.join(
            self.url,
            'vnflcm/v1/vnf_lcm_op_occs',
            uuidsentinel.vnf_lcm_op_occ_id,
            'cancel')
        self.requests_mock.register_uri(
            'POST', url, headers=self.header, status_code=404, json={})

        self.assertRaises(exceptions.TackerClientException,
                          self.cancel_vnf_lcm.take_action,
                          parsed_args)


class TestRollbackVnfLcmOp(TestVnfLcm):

    def setUp(self):
        super(TestRollbackVnfLcmOp, self).setUp()
        self.rollback_vnf_lcm = vnflcm_op_occs.RollbackVnfLcmOp(
            self.app, self.app_args, cmd_name='vnflcm op rollback')

    def test_take_action(self):
        """take_action normal system test"""

        arglist = [uuidsentinel.vnf_lcm_op_occ_id]
        verifylist = [('vnf_lcm_op_occ_id', uuidsentinel.vnf_lcm_op_occ_id)]

        parsed_args = self.check_parser(
            self.rollback_vnf_lcm, arglist, verifylist)
        url = os.path.join(
            self.url,
            'vnflcm/v1/vnf_lcm_op_occs',
            uuidsentinel.vnf_lcm_op_occ_id,
            'rollback')

        self.requests_mock.register_uri(
            'POST', url, headers=self.header, json={})

        sys.stdout = buffer = StringIO()
        self.rollback_vnf_lcm.take_action(parsed_args)

        actual_message = buffer.getvalue().strip()

        expected_message = (
            'Rollback request for LCM operation ' +
            uuidsentinel.vnf_lcm_op_occ_id +
            ' has been accepted')

        self.assertEqual(expected_message, actual_message)

    def test_take_action_vnf_lcm_op_occ_id_not_found(self):
        """take_action abnomaly system test"""

        arglist = [uuidsentinel.vnf_lcm_op_occ_id]
        verifylist = [('vnf_lcm_op_occ_id', uuidsentinel.vnf_lcm_op_occ_id)]

        parsed_args = self.check_parser(
            self.rollback_vnf_lcm, arglist, verifylist)

        url = os.path.join(
            self.url,
            'vnflcm/v1/vnf_lcm_op_occs',
            uuidsentinel.vnf_lcm_op_occ_id,
            'rollback')
        self.requests_mock.register_uri(
            'POST', url, headers=self.header, status_code=404, json={})

        self.assertRaises(exceptions.TackerClientException,
                          self.rollback_vnf_lcm.take_action,
                          parsed_args)


class TestFailVnfLcmOp(TestVnfLcm):

    def setUp(self):
        super(TestFailVnfLcmOp, self).setUp()
        self.fail_vnf_lcm = vnflcm_op_occs.FailVnfLcmOp(
            self.app, self.app_args, cmd_name='vnflcm op fail')

    def test_take_action(self):
        """Test of take_action()"""

        vnflcm_op_occ = vnflcm_op_occs_fakes.vnflcm_op_occ_response(
            action='fail')

        arg_list = [vnflcm_op_occ['id']]
        verify_list = [('vnf_lcm_op_occ_id', vnflcm_op_occ['id'])]

        # command param
        parsed_args = self.check_parser(
            self.fail_vnf_lcm, arg_list, verify_list)
        url = os.path.join(
            self.url,
            'vnflcm/v1/vnf_lcm_op_occs',
            vnflcm_op_occ['id'],
            'fail')

        self.requests_mock.register_uri(
            'POST', url, headers=self.header, json=vnflcm_op_occ)

        columns, data = (self.fail_vnf_lcm.take_action(parsed_args))
        expected_columns = _get_columns_vnflcm_op_occs(action='fail')

        self.assertCountEqual(expected_columns, columns)

    def test_take_action_vnf_lcm_op_occ_id_not_found(self):
        """Test if vnf-lcm-op-occ-id does not find"""

        arg_list = [uuidsentinel.vnf_lcm_op_occ_id]
        verify_list = [('vnf_lcm_op_occ_id', uuidsentinel.vnf_lcm_op_occ_id)]

        # command param
        parsed_args = self.check_parser(
            self.fail_vnf_lcm, arg_list, verify_list)

        url = os.path.join(
            self.url,
            'vnflcm/v1/vnf_lcm_op_occs',
            uuidsentinel.vnf_lcm_op_occ_id,
            'fail')
        self.requests_mock.register_uri(
            'POST', url, headers=self.header, status_code=404, json={})

        self.assertRaises(exceptions.TackerClientException,
                          self.fail_vnf_lcm.take_action,
                          parsed_args)

    def test_take_action_vnf_lcm_op_occ_state_is_conflict(self):
        """Test if vnf-lcm-op-occ state is conflict"""

        arg_list = [uuidsentinel.vnf_lcm_op_occ_id]
        verify_list = [('vnf_lcm_op_occ_id', uuidsentinel.vnf_lcm_op_occ_id)]

        # command param
        parsed_args = self.check_parser(
            self.fail_vnf_lcm, arg_list, verify_list)

        url = os.path.join(
            self.url,
            'vnflcm/v1/vnf_lcm_op_occs',
            uuidsentinel.vnf_lcm_op_occ_id,
            'fail')
        self.requests_mock.register_uri(
            'POST', url, headers=self.header, status_code=409, json={})

        self.assertRaises(exceptions.TackerClientException,
                          self.fail_vnf_lcm.take_action,
                          parsed_args)

    def test_take_action_vnf_lcm_op_occ_internal_server_error(self):
        """Test if request is internal server error"""

        arg_list = [uuidsentinel.vnf_lcm_op_occ_id]
        verify_list = [('vnf_lcm_op_occ_id', uuidsentinel.vnf_lcm_op_occ_id)]

        # command param
        parsed_args = self.check_parser(
            self.fail_vnf_lcm, arg_list, verify_list)

        url = os.path.join(
            self.url,
            'vnflcm/v1/vnf_lcm_op_occs',
            uuidsentinel.vnf_lcm_op_occ_id,
            'fail')
        self.requests_mock.register_uri(
            'POST', url, headers=self.header, status_code=500, json={})

        self.assertRaises(exceptions.TackerClientException,
                          self.fail_vnf_lcm.take_action,
                          parsed_args)

    def test_take_action_vnf_lcm_op_occ_missing_vnf_lcm_op_occ_id_argument(
        self):
        """Test if vnflcm_op_occ_id is not provided"""

        arg_list = []
        verify_list = [('vnf_lcm_op_occ_id', arg_list)]
        self.assertRaises(base.ParserException, self.check_parser,
                          self.fail_vnf_lcm, arg_list, verify_list)


class TestRetryVnfLcmOp(TestVnfLcm):

    def setUp(self):
        super(TestRetryVnfLcmOp, self).setUp()
        self.retry_vnf_lcm = vnflcm_op_occs.RetryVnfLcmOp(
            self.app, self.app_args, cmd_name='vnflcm op retry')

    def test_take_action(self):
        """Test of take_action()"""
        arg_list = [uuidsentinel.vnf_lcm_op_occ_id]
        verify_list = [('vnf_lcm_op_occ_id', uuidsentinel.vnf_lcm_op_occ_id)]

        # command param
        parsed_args = self.check_parser(
            self.retry_vnf_lcm, arg_list, verify_list)
        url = os.path.join(
            self.url,
            'vnflcm/v1/vnf_lcm_op_occs',
            uuidsentinel.vnf_lcm_op_occ_id,
            'retry')

        self.requests_mock.register_uri(
            'POST', url, headers=self.header, json={})

        sys.stdout = buffer = StringIO()
        self.retry_vnf_lcm.take_action(parsed_args)

        actual_message = buffer.getvalue().strip()

        expected_message = (
            'Retry request for LCM operation ' +
            uuidsentinel.vnf_lcm_op_occ_id +
            ' has been accepted')

        self.assertEqual(expected_message, actual_message)

    def test_take_action_vnf_lcm_op_occ_id_not_found(self):
        """Test if vnf-lcm-op-occ-id is not found."""
        arglist = [uuidsentinel.vnf_lcm_op_occ_id]
        verifylist = [('vnf_lcm_op_occ_id', uuidsentinel.vnf_lcm_op_occ_id)]

        # command param
        parsed_args = self.check_parser(
            self.retry_vnf_lcm, arglist, verifylist)

        url = os.path.join(
            self.url,
            'vnflcm/v1/vnf_lcm_op_occs',
            uuidsentinel.vnf_lcm_op_occ_id,
            'retry')

        self.requests_mock.register_uri(
            'POST', url, headers=self.header, status_code=404, json={})

        self.assertRaises(exceptions.TackerClientException,
                          self.retry_vnf_lcm.take_action,
                          parsed_args)

    def test_take_action_vnf_lcm_op_occ_state_is_conflict(self):
        """Test if vnf-lcm-op-occ state is conflict"""

        arg_list = [uuidsentinel.vnf_lcm_op_occ_id]
        verify_list = [('vnf_lcm_op_occ_id', uuidsentinel.vnf_lcm_op_occ_id)]

        # command param
        parsed_args = self.check_parser(
            self.retry_vnf_lcm, arg_list, verify_list)

        url = os.path.join(
            self.url,
            'vnflcm/v1/vnf_lcm_op_occs',
            uuidsentinel.vnf_lcm_op_occ_id,
            'retry')

        self.requests_mock.register_uri(
            'POST', url, headers=self.header, status_code=409, json={})

        self.assertRaises(exceptions.TackerClientException,
                          self.retry_vnf_lcm.take_action,
                          parsed_args)

    def test_take_action_vnf_lcm_op_occ_internal_server_error(self):
        """Test if request is internal server error"""

        arg_list = [uuidsentinel.vnf_lcm_op_occ_id]
        verify_list = [('vnf_lcm_op_occ_id', uuidsentinel.vnf_lcm_op_occ_id)]

        # command param
        parsed_args = self.check_parser(
            self.retry_vnf_lcm, arg_list, verify_list)

        url = os.path.join(
            self.url,
            'vnflcm/v1/vnf_lcm_op_occs',
            uuidsentinel.vnf_lcm_op_occ_id,
            'retry')

        self.requests_mock.register_uri(
            'POST', url, headers=self.header, status_code=500, json={})

        self.assertRaises(exceptions.TackerClientException,
                          self.retry_vnf_lcm.take_action,
                          parsed_args)

    def test_take_action_vnf_lcm_op_occ_missing_vnf_lcm_op_occ_id_argument(
        self):
        """Test if vnflcm_op_occ_id is not provided"""

        arg_list = []
        verify_list = [('vnf_lcm_op_occ_id', arg_list)]
        self.assertRaises(base.ParserException, self.check_parser,
                          self.retry_vnf_lcm, arg_list, verify_list)


class TestListVnfLcmOp(TestVnfLcm):

    def setUp(self):
        super(TestListVnfLcmOp, self).setUp()
        self.list_vnflcm_op_occ = vnflcm_op_occs.ListVnfLcmOp(
            self.app, self.app_args, cmd_name='vnflcm op list')

    def test_take_action(self):
        vnflcm_op_occs_obj = vnflcm_op_occs_fakes.create_vnflcm_op_occs(
            count=3)
        parsed_args = self.check_parser(self.list_vnflcm_op_occ, [], [])
        self.requests_mock.register_uri(
            'GET', os.path.join(self.url,
                                'vnflcm/v1/vnf_lcm_op_occs'),
            json=vnflcm_op_occs_obj, headers=self.header)

        actual_columns, data = self.list_vnflcm_op_occ.take_action(parsed_args)

        headers, columns = tacker_osc_utils.get_column_definitions(
            self.list_vnflcm_op_occ.get_attributes(), long_listing=True)

        expected_data = []
        for vnflcm_op_occ_obj_idx in vnflcm_op_occs_obj:
            expected_data.append(vnflcm_op_occs_fakes.get_vnflcm_op_occ_data(
                vnflcm_op_occ_obj_idx, columns=columns))

        self.assertCountEqual(_get_columns_vnflcm_op_occs(action='list'),
                              actual_columns)
        self.assertCountEqual(expected_data, list(data))

    def test_take_action_with_filter(self):
        vnflcm_op_occs_obj = vnflcm_op_occs_fakes.create_vnflcm_op_occs(
            count=3)
        parsed_args = self.check_parser(
            self.list_vnflcm_op_occ,
            ["--filter", '(eq,operationState,STARTING)'],
            [('filter', '(eq,operationState,STARTING)')])
        self.requests_mock.register_uri(
            'GET', os.path.join(
                self.url,
                'vnflcm/v1/vnf_lcm_op_occs?'
                'filter=(eq,operationState,STARTING)'),
            json=vnflcm_op_occs_obj, headers=self.header)

        actual_columns, data = self.list_vnflcm_op_occ.take_action(parsed_args)

        headers, columns = tacker_osc_utils.get_column_definitions(
            self.list_vnflcm_op_occ.get_attributes(), long_listing=True)

        expected_data = []
        for vnflcm_op_occ_obj_idx in vnflcm_op_occs_obj:
            expected_data.append(vnflcm_op_occs_fakes.get_vnflcm_op_occ_data(
                vnflcm_op_occ_obj_idx, columns=columns))

        self.assertCountEqual(_get_columns_vnflcm_op_occs(action='list'),
                              actual_columns)
        self.assertListItemsEqual(expected_data, list(data))

    def test_take_action_with_incorrect_filter(self):

        parsed_args = self.check_parser(
            self.list_vnflcm_op_occ,
            ["--filter", '(operationState)'],
            [('filter', '(operationState)')])

        url = os.path.join(
            self.url,
            'vnflcm/v1/vnf_lcm_op_occs?filter=(operationState)')
        self.requests_mock.register_uri(
            'POST', url, headers=self.header, status_code=400, json={})

        self.assertRaises(exceptions.TackerClientException,
                          self.list_vnflcm_op_occ.take_action,
                          parsed_args)

    def test_take_action_internal_server_error(self):

        parsed_args = self.check_parser(
            self.list_vnflcm_op_occ,
            ["--filter", '(eq,operationState,STARTING)'],
            [('filter', '(eq,operationState,STARTING)')])

        url = os.path.join(
            self.url,
            'vnflcm/v1/vnf_lcm_op_occs?'
            'filter=(eq,operationState,STARTING)')
        self.requests_mock.register_uri(
            'POST', url, headers=self.header, status_code=500, json={})

        self.assertRaises(exceptions.TackerClientException,
                          self.list_vnflcm_op_occ.take_action,
                          parsed_args)

    def test_take_action_with_exclude_fields(self):

        vnflcm_op_occs_obj = vnflcm_op_occs_fakes.create_vnflcm_op_occs(
            count=3)
        parsed_args = self.check_parser(
            self.list_vnflcm_op_occ,
            ["--exclude-fields", 'VNF Instance ID,Operation State'],
            [('exclude_fields', 'VNF Instance ID,Operation State')])
        self.requests_mock.register_uri(
            'GET', os.path.join(
                self.url,
                'vnflcm/v1/vnf_lcm_op_occs?'
                'exclude-fields=VNF Instance ID,Operation State'),
            json=vnflcm_op_occs_obj, headers=self.header)
        actual_columns, data = self.list_vnflcm_op_occ.take_action(parsed_args)
        headers, columns = tacker_osc_utils.get_column_definitions(
            self.list_vnflcm_op_occ.get_attributes(
                exclude=['VNF Instance ID', 'Operation State']),
            long_listing=True)
        expected_data = []
        for vnflcm_op_occ_obj_idx in vnflcm_op_occs_obj:
            expected_data.append(
                vnflcm_op_occs_fakes.get_vnflcm_op_occ_data(
                    vnflcm_op_occ_obj_idx, columns=columns))

        self.assertCountEqual(_get_columns_vnflcm_op_occs(
            action='list', parameter="exclude_fields"),
            actual_columns)
        self.assertListItemsEqual(expected_data, list(data))

    def test_take_action_with_pagination(self):
        next_links_num = 3
        vnflcm_op_occs_obj = vnflcm_op_occs_fakes.create_vnflcm_op_occs(
            count=next_links_num)
        parsed_args = self.check_parser(self.list_vnflcm_op_occ, [], [])
        path = os.path.join(self.url, 'vnflcm/v1/vnf_lcm_op_occs')

        links = [0] * next_links_num
        link_headers = [0] * next_links_num

        for i in range(next_links_num):
            links[i] = (
                '{base_url}?nextpage_opaque_marker={vnflcm_op_occ_id}'.format(
                    base_url=path,
                    vnflcm_op_occ_id=vnflcm_op_occs_obj[i]['id']))
            link_headers[i] = copy.deepcopy(self.header)
            link_headers[i]['Link'] = '<{link_url}>; rel="next"'.format(
                link_url=links[i])

        self.requests_mock.register_uri(
            'GET', path, json=[vnflcm_op_occs_obj[0]], headers=link_headers[0])
        self.requests_mock.register_uri(
            'GET', links[0], json=[vnflcm_op_occs_obj[1]],
            headers=link_headers[1])
        self.requests_mock.register_uri(
            'GET', links[1], json=[vnflcm_op_occs_obj[2]],
            headers=link_headers[2])
        self.requests_mock.register_uri(
            'GET', links[2], json=[], headers=self.header)

        actual_columns, data = self.list_vnflcm_op_occ.take_action(parsed_args)

        headers, columns = tacker_osc_utils.get_column_definitions(
            self.list_vnflcm_op_occ.get_attributes(), long_listing=True)

        expected_data = []
        for vnflcm_op_occ_obj_idx in vnflcm_op_occs_obj:
            expected_data.append(vnflcm_op_occs_fakes.get_vnflcm_op_occ_data(
                vnflcm_op_occ_obj_idx, columns=columns))

        self.assertCountEqual(_get_columns_vnflcm_op_occs(action='list'),
                              actual_columns)
        self.assertCountEqual(expected_data, list(data))


class TestShowVnfLcmOp(TestVnfLcm):

    def setUp(self):
        super(TestShowVnfLcmOp, self).setUp()
        self.show_vnf_lcm_op_occs = vnflcm_op_occs.ShowVnfLcmOp(
            self.app, self.app_args, cmd_name='vnflcm op show')

    def test_take_action(self):
        """Test of take_action()"""
        vnflcm_op_occ = vnflcm_op_occs_fakes.vnflcm_op_occ_response()

        arglist = [vnflcm_op_occ['id']]
        verifylist = [('vnf_lcm_op_occ_id', vnflcm_op_occ['id'])]

        # command param
        parsed_args = self.check_parser(
            self.show_vnf_lcm_op_occs, arglist, verifylist)
        url = os.path.join(
            self.url,
            'vnflcm/v1/vnf_lcm_op_occs',
            vnflcm_op_occ['id'])

        self.requests_mock.register_uri(
            'GET', url, headers=self.header, json=vnflcm_op_occ)

        columns, data = (self.show_vnf_lcm_op_occs.take_action(parsed_args))

        self.assertCountEqual(_get_columns_vnflcm_op_occs(),
                              columns)

    def test_take_action_vnf_lcm_op_occ_id_not_found(self):
        """Test if vnf-lcm-op-occ-id does not find."""
        arglist = [uuidsentinel.vnf_lcm_op_occ_id]
        verifylist = [('vnf_lcm_op_occ_id', uuidsentinel.vnf_lcm_op_occ_id)]

        # command param
        parsed_args = self.check_parser(
            self.show_vnf_lcm_op_occs, arglist, verifylist)

        url = os.path.join(
            self.url,
            'vnflcm/v1/vnf_lcm_op_occs',
            uuidsentinel.vnf_lcm_op_occ_id)
        self.requests_mock.register_uri(
            'GET', url, headers=self.header, status_code=404, json={})

        self.assertRaises(exceptions.TackerClientException,
                          self.show_vnf_lcm_op_occs.take_action,
                          parsed_args)

    def test_take_action_internal_server_error(self):
        """Test for internal server error."""
        arglist = [uuidsentinel.vnf_lcm_op_occ_id]
        verifylist = [('vnf_lcm_op_occ_id', uuidsentinel.vnf_lcm_op_occ_id)]

        # command param
        parsed_args = self.check_parser(
            self.show_vnf_lcm_op_occs, arglist, verifylist)

        url = os.path.join(
            self.url,
            'vnflcm/v1/vnf_lcm_op_occs',
            uuidsentinel.vnf_lcm_op_occ_id)
        self.requests_mock.register_uri(
            'GET', url, headers=self.header, status_code=500, json={})

        self.assertRaises(exceptions.TackerClientException,
                          self.show_vnf_lcm_op_occs.take_action,
                          parsed_args)



# Copyright (C) 2020 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import time

from osc_lib.command import command
from osc_lib import utils

from tackerclient.common import exceptions
from tackerclient.i18n import _
from tackerclient.osc import sdk_utils
from tackerclient.osc import utils as tacker_osc_utils

_attr_map = (
    ('id', 'ID', tacker_osc_utils.LIST_BOTH),
    ('vnfInstanceName', 'VNF Instance Name', tacker_osc_utils.LIST_BOTH),
    ('instantiationState', 'Instantiation State', tacker_osc_utils.LIST_BOTH),
    ('vnfProvider', 'VNF Provider', tacker_osc_utils.LIST_BOTH),
    ('vnfSoftwareVersion', 'VNF Software Version', tacker_osc_utils.LIST_BOTH),
    ('vnfProductName', 'VNF Product Name', tacker_osc_utils.LIST_BOTH),
    ('vnfdId', 'VNFD ID', tacker_osc_utils.LIST_BOTH)
)

LOG = logging.getLogger(__name__)

_mixed_case_fields = ('vnfInstanceName', 'vnfInstanceDescription', 'vnfdId',
                      'vnfProvider', 'vnfProductName', 'vnfSoftwareVersion',
                      'vnfdVersion', 'instantiationState',
                      'vimConnectionInfo', 'instantiatedVnfInfo',
                      'vnfConfigurableProperties', 'vnfPkgId')

_VNF_INSTANCE = 'vnf_instance'

VNF_INSTANCE_TERMINATION_TIMEOUT = 300

EXTRA_WAITING_TIME = 10

SLEEP_TIME = 1

formatters = {'vimConnectionInfo': tacker_osc_utils.FormatComplexDataColumn,
              'instantiatedVnfInfo': tacker_osc_utils.FormatComplexDataColumn,
              '_links': tacker_osc_utils.FormatComplexDataColumn}


def _get_columns(vnflcm_obj, action=None):
    column_map = {
        'id': 'ID',
        'vnfInstanceName': 'VNF Instance Name',
        'vnfInstanceDescription': 'VNF Instance Description',
        'vnfdId': 'VNFD ID',
        'vnfProvider': 'VNF Provider',
        'vnfProductName': 'VNF Product Name',
        'vnfSoftwareVersion': 'VNF Software Version',
        'vnfdVersion': 'VNFD Version',
        'instantiationState': 'Instantiation State',
        '_links': 'Links',
        'vnfConfigurableProperties': 'VNF Configurable Properties'
    }
    if action == 'show':
        if vnflcm_obj['instantiationState'] == 'INSTANTIATED':
            column_map.update(
                {'instantiatedVnfInfo': 'Instantiated Vnf Info'}
            )
        column_map.update(
            {'vimConnectionInfo': 'VIM Connection Info',
             '_links': 'Links'}
        )
    # Note: To prevent it from appearing in the v2 API output,
    # the 'VNF Package ID' will be output only if the vnfPkgId exists.
    if 'vnfPkgId' in vnflcm_obj:
        column_map.update(
            {'vnfPkgId': 'VNF Package ID'}
        )
    return sdk_utils.get_osc_show_columns_for_sdk_resource(vnflcm_obj,
                                                           column_map)


class CreateVnfLcm(command.ShowOne):
    _description = _("Create a new VNF Instance")

    def get_parser(self, prog_name):
        parser = super(CreateVnfLcm, self).get_parser(prog_name)
        parser.add_argument(
            'vnfd_id',
            metavar="<vnfd-id>",
            help=_('Identifier that identifies the VNFD which defines the '
                   'VNF instance to be created.'))
        parser.add_argument(
            '--name',
            metavar="<vnf-instance-name>",
            help=_('Name of the VNF instance to be created.'))
        parser.add_argument(
            '--description',
            metavar="<vnf-instance-description>",
            help=_('Description of the VNF instance to be created.'))
        parser.add_argument(
            '--I',
            metavar="<param-file>",
            help=_("Instantiate VNF subsequently after it's creation. "
                   "Specify instantiate request parameters in a json file."))
        return parser

    def args2body(self, parsed_args, file_path=None):
        body = {}

        if file_path:
            return tacker_osc_utils.jsonfile2body(file_path)

        body['vnfdId'] = parsed_args.vnfd_id

        if parsed_args.description:
            body['vnfInstanceDescription'] = parsed_args.description

        if parsed_args.name:
            body['vnfInstanceName'] = parsed_args.name

        return body

    def take_action(self, parsed_args):
        client = self.app.client_manager.tackerclient
        vnf = client.create_vnf_instance(self.args2body(parsed_args))
        if parsed_args.I:
            # Instantiate VNF instance.
            result = client.instantiate_vnf_instance(
                vnf['id'],
                self.args2body(parsed_args, file_path=parsed_args.I))
            if not result:
                print((_('VNF Instance %(id)s is created and instantiation'
                         ' request has been accepted.') % {'id': vnf['id']}))
        display_columns, columns = _get_columns(vnf)
        data = utils.get_item_properties(sdk_utils.DictModel(vnf),
                                         columns, formatters=formatters,
                                         mixed_case_fields=_mixed_case_fields)
        return (display_columns, data)


class ShowVnfLcm(command.ShowOne):
    _description = _("Display VNF instance details")

    def get_parser(self, prog_name):
        parser = super(ShowVnfLcm, self).get_parser(prog_name)
        parser.add_argument(
            _VNF_INSTANCE,
            metavar="<vnf-instance>",
            help=_("VNF instance ID to display"))
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.tackerclient
        obj = client.show_vnf_instance(parsed_args.vnf_instance)
        display_columns, columns = _get_columns(obj, action='show')
        data = utils.get_item_properties(
            sdk_utils.DictModel(obj),
            columns, mixed_case_fields=_mixed_case_fields,
            formatters=formatters)
        return (display_columns, data)


class ListVnfLcm(command.Lister):
    _description = _("List VNF Instance")

    def get_parser(self, prog_name):
        parser = super(ListVnfLcm, self).get_parser(prog_name)
        return parser

    def take_action(self, parsed_args):
        _params = {}
        client = self.app.client_manager.tackerclient
        vnf_instances = client.list_vnf_instances(**_params)
        headers, columns = tacker_osc_utils.get_column_definitions(
            _attr_map, long_listing=True)
        return (headers,
                (utils.get_dict_properties(
                    s, columns, mixed_case_fields=_mixed_case_fields,
                ) for s in vnf_instances))


class InstantiateVnfLcm(command.Command):
    _description = _("Instantiate a VNF Instance")

    def get_parser(self, prog_name):
        parser = super(InstantiateVnfLcm, self).get_parser(prog_name)
        parser.add_argument(
            _VNF_INSTANCE,
            metavar="<vnf-instance>",
            help=_("VNF instance ID to instantiate"))
        parser.add_argument(
            'instantiation_request_file',
            metavar="<param-file>",
            help=_('Specify instantiate request parameters in a json file.'))

        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.tackerclient
        result = client.instantiate_vnf_instance(
            parsed_args.vnf_instance, tacker_osc_utils.jsonfile2body(
                parsed_args.instantiation_request_file))
        if not result:
            print((_('Instantiate request for VNF Instance %(id)s has been'
                     ' accepted.') % {'id': parsed_args.vnf_instance}))


class HealVnfLcm(command.Command):
    _description = _("Heal VNF Instance")

    def get_parser(self, prog_name):
        parser = super(HealVnfLcm, self).get_parser(prog_name)
        usage_message = ('''%(prog)s [-h] [--cause CAUSE]
                             [--vnfc-instance <vnfc-instance-id> '''
                         '''[<vnfc-instance-id> ...]]
                             [--additional-param-file <additional-param-file>]
                             -- <vnf-instance>''')
        parser.usage = usage_message
        parser.add_argument(
            _VNF_INSTANCE,
            metavar="<vnf-instance>",
            help=_("VNF instance ID to heal"))
        parser.add_argument(
            '--cause',
            help=_('Specify the reason why a healing procedure is required.'))
        parser.add_argument(
            '--vnfc-instance',
            metavar="<vnfc-instance-id>",
            nargs="+",
            help=_("List of VNFC instances requiring a healing action.")
        )
        parser.add_argument(
            '--additional-param-file',
            metavar="<additional-param-file>",
            help=_("Additional parameters passed by the NFVO as input "
                   "to the healing process."))
        return parser

    def args2body(self, parsed_args):
        body = {}
        if parsed_args.cause:
            body['cause'] = parsed_args.cause
        if parsed_args.vnfc_instance:
            body['vnfcInstanceId'] = parsed_args.vnfc_instance
        if parsed_args.additional_param_file:
            body.update(tacker_osc_utils.jsonfile2body(
                parsed_args.additional_param_file))

        return body

    def take_action(self, parsed_args):
        client = self.app.client_manager.tackerclient
        result = client.heal_vnf_instance(
            parsed_args.vnf_instance, self.args2body(parsed_args))
        if not result:
            print((_('Heal request for VNF Instance %(id)s has been'
                     ' accepted.') % {'id': parsed_args.vnf_instance}))


class TerminateVnfLcm(command.Command):
    _description = _("Terminate a VNF instance")

    def get_parser(self, prog_name):
        parser = super(TerminateVnfLcm, self).get_parser(prog_name)
        parser.add_argument(
            _VNF_INSTANCE,
            metavar="<vnf-instance>",
            help=_("VNF instance ID to terminate"))
        parser.add_argument(
            "--termination-type",
            default='GRACEFUL',
            metavar="<termination-type>",
            choices=['GRACEFUL', 'FORCEFUL'],
            help=_("Termination type can be 'GRACEFUL' or 'FORCEFUL'. "
                   "Default is 'GRACEFUL'"))
        parser.add_argument(
            '--graceful-termination-timeout',
            metavar="<graceful-termination-timeout>",
            type=int,
            help=_('This attribute is only applicable in case of graceful '
                   'termination. It defines the time to wait for the VNF to be'
                   ' taken out of service before shutting down the VNF and '
                   'releasing the resources. The unit is seconds.'))
        parser.add_argument(
            '--D',
            action='store_true',
            default=False,
            help=_("Delete VNF Instance subsequently after it's termination"),
        )
        return parser

    def args2body(self, parsed_args):
        body = {}
        body['terminationType'] = parsed_args.termination_type

        if parsed_args.graceful_termination_timeout:
            if parsed_args.termination_type == 'FORCEFUL':
                exceptions.InvalidInput(reason='--graceful-termination-timeout'
                                        ' argument is invalid for "FORCEFUL"'
                                        ' termination')
            body['gracefulTerminationTimeout'] = parsed_args.\
                graceful_termination_timeout

        return body

    def take_action(self, parsed_args):
        client = self.app.client_manager.tackerclient
        result = client.terminate_vnf_instance(parsed_args.vnf_instance,
                                               self.args2body(parsed_args))
        if not result:
            print(_("Terminate request for VNF Instance '%(id)s' has been"
                  " accepted.") % {'id': parsed_args.vnf_instance})
            if parsed_args.D:
                print(_("Waiting for vnf instance to be terminated before "
                      "deleting"))

                self._wait_until_vnf_is_terminated(
                    client, parsed_args.vnf_instance,
                    graceful_timeout=parsed_args.graceful_termination_timeout)

                result = client.delete_vnf_instance(parsed_args.vnf_instance)
                if not result:
                    print(_("VNF Instance '%(id)s' is deleted successfully") %
                          {'id': parsed_args.vnf_instance})

    def _wait_until_vnf_is_terminated(self, client, vnf_instance_id,
                                      graceful_timeout=None):
        # wait until vnf instance 'instantiationState' is set to
        # 'NOT_INSTANTIATED'
        if graceful_timeout:
            # If graceful_termination_timeout is provided,
            # terminate vnf will start after this timeout period.
            # Hence, it should wait for extra time of 10 seconds
            # after this graceful_termination_timeout period.
            timeout = graceful_timeout + EXTRA_WAITING_TIME
        else:
            timeout = VNF_INSTANCE_TERMINATION_TIMEOUT

        start_time = int(time.time())
        while True:
            vnf_instance = client.show_vnf_instance(vnf_instance_id)
            if vnf_instance['instantiationState'] == 'NOT_INSTANTIATED':
                break

            if ((int(time.time()) - start_time) > timeout):
                msg = _("Couldn't verify vnf instance is terminated within "
                        "'%(timeout)s' seconds. Unable to delete vnf instance "
                        "%(id)s")
                raise exceptions.CommandError(
                    message=msg % {'timeout': timeout, 'id': vnf_instance_id})
            time.sleep(SLEEP_TIME)


class DeleteVnfLcm(command.Command):
    """Vnf lcm delete

    DeleteVnfLcm class supports bulk deletion of vnf instances, and error
    handling.
    """

    _description = _("Delete VNF Instance(s)")

    def get_parser(self, prog_name):
        parser = super(DeleteVnfLcm, self).get_parser(prog_name)
        parser.add_argument(
            'vnf_instances',
            metavar="<vnf-instance>",
            nargs="+",
            help=_("VNF instance ID(s) to delete"))
        return parser

    def take_action(self, parsed_args):
        error_count = 0
        client = self.app.client_manager.tackerclient
        vnf_instances = parsed_args.vnf_instances
        for vnf_instance in vnf_instances:
            try:
                client.delete_vnf_instance(vnf_instance)
            except Exception as e:
                error_count += 1
                LOG.error(_("Failed to delete vnf instance with "
                            "ID '%(vnf)s': %(e)s"),
                          {'vnf': vnf_instance, 'e': e})

        total = len(vnf_instances)
        if (error_count > 0):
            msg = (_("Failed to delete %(error_count)s of %(total)s "
                     "vnf instances.") % {'error_count': error_count,
                                          'total': total})
            raise exceptions.CommandError(message=msg)
        else:
            if total > 1:
                print(_('All specified vnf instances are deleted '
                        'successfully'))
            else:
                print(_("Vnf instance '%s' is deleted "
                        "successfully") % vnf_instances[0])


class UpdateVnfLcm(command.Command):
    _description = _("Update VNF Instance")

    def get_parser(self, prog_name):
        """Add arguments to parser.

        Args:
            prog_name ([string]): program name

        Returns:
            parser([ArgumentParser]): [description]
        """
        parser = super(UpdateVnfLcm, self).get_parser(prog_name)
        parser.add_argument(
            _VNF_INSTANCE,
            metavar="<vnf-instance>",
            help=_('VNF instance ID to update.'))
        parser.add_argument(
            '--I',
            metavar="<param-file>",
            help=_("Specify update request parameters in a json file."))

        return parser

    def args2body(self, file_path=None):
        """Call jsonfile2body to store request body to body(dict)

        Args:
            file_path ([string], optional): file path of param file(json).
                                             Defaults to None.

        Returns:
            body ([dict]): Request body is stored
        """
        body = {}

        if file_path:
            return tacker_osc_utils.jsonfile2body(file_path)

        return body

    def take_action(self, parsed_args):
        """Execute update_vnf_instance and output result comment

        Args:
            parsed_args ([Namespace]): [description]
        """
        client = self.app.client_manager.tackerclient
        if parsed_args.I:
            # Update VNF instance.
            result = client.update_vnf_instance(
                parsed_args.vnf_instance,
                self.args2body(file_path=parsed_args.I))
            if not result:
                print((_('Update vnf:%(id)s ') %
                       {'id': parsed_args.vnf_instance}))


class ScaleVnfLcm(command.Command):
    _description = _("Scale a VNF Instance")

    def get_parser(self, prog_name):
        parser = super(ScaleVnfLcm, self).get_parser(prog_name)
        parser.add_argument(
            _VNF_INSTANCE,
            metavar="<vnf-instance>",
            help=_('VNF instance ID to scale'))
        parser.add_argument(
            '--number-of-steps',
            metavar="<number-of-steps>",
            type=int,
            help=_("Number of scaling steps to be executed as part of "
                   "this Scale VNF operation."))
        parser.add_argument(
            '--additional-param-file',
            metavar="<additional-param-file>",
            help=_("Additional parameters passed by the NFVO as input "
                   "to the scaling process."))

        scale_require_parameters = parser.add_argument_group(
            "require arguments"
        )
        scale_require_parameters.add_argument(
            '--type',
            metavar="<type>",
            required=True,
            choices=['SCALE_OUT', 'SCALE_IN'],
            help=_("SCALE_OUT or SCALE_IN for type of scale operation."))
        scale_require_parameters.add_argument(
            '--aspect-id',
            required=True,
            metavar="<aspect-id>",
            help=_("Identifier of the scaling aspect."))

        return parser

    def args2body(self, parsed_args):
        """To store request body, call jsonfile2body.

        Args:
            parsed_args ([Namespace]): arguments of CLI.

        Returns:
            body ([dict]): Request body is stored
        """
        body = {'type': parsed_args.type, 'aspectId': parsed_args.aspect_id}

        if parsed_args.number_of_steps:
            body['numberOfSteps'] = parsed_args.number_of_steps

        if parsed_args.additional_param_file:
            body.update(tacker_osc_utils.jsonfile2body(
                parsed_args.additional_param_file))

        return body

    def take_action(self, parsed_args):
        """Execute scale_vnf_instance and output result comment.

        Args:
            parsed_args ([Namespace]): arguments of CLI.
        """
        client = self.app.client_manager.tackerclient
        result = client.scale_vnf_instance(
            parsed_args.vnf_instance,
            self.args2body(parsed_args))
        if not result:
            print((_('Scale request for VNF Instance %s has been accepted.')
                   % parsed_args.vnf_instance))


class ChangeExtConnVnfLcm(command.Command):
    _description = _("Change External VNF Connectivity")

    def get_parser(self, prog_name):
        parser = super(ChangeExtConnVnfLcm, self).get_parser(prog_name)
        parser.add_argument(
            _VNF_INSTANCE,
            metavar="<vnf-instance>",
            help=_("VNF instance ID to Change External VNF Connectivity"))
        parser.add_argument(
            'request_file',
            metavar="<param-file>",
            help=_("Specify change-ext-conn request parameters "
                   "in a json file."))

        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.tackerclient
        result = client.change_ext_conn_vnf_instance(
            parsed_args.vnf_instance, tacker_osc_utils.jsonfile2body(
                parsed_args.request_file))
        if not result:
            print((_('Change External VNF Connectivity for VNF Instance %s '
                     'has been accepted.') % parsed_args.vnf_instance))


class ChangeVnfPkgVnfLcm(command.Command):
    _description = _("Change Current VNF Package")

    def get_parser(self, prog_name):
        parser = super(ChangeVnfPkgVnfLcm, self).get_parser(prog_name)
        parser.add_argument(
            _VNF_INSTANCE,
            metavar="<vnf-instance>",
            help=_("VNF instance ID to Change Current VNF Package"))
        parser.add_argument(
            'request_file',
            metavar="<param-file>",
            help=_("Specify change-vnfpkg request parameters "
                   "in a json file."))

        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.tackerclient
        result = client.change_vnfpkg_vnf_instance(
            parsed_args.vnf_instance, tacker_osc_utils.jsonfile2body(
                parsed_args.request_file))
        if not result:
            print((_('Change Current VNF Package for VNF Instance %s '
                     'has been accepted.') % parsed_args.vnf_instance))



# Copyright 2012 OpenStack Foundation.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from tackerclient.common._i18n import _
from tackerclient.common import exceptions
from tackerclient.common import utils


API_NAME = 'nfv-orchestration'
API_VERSIONS = {
    '1.0': 'tackerclient.v1_0.client.Client',
}


def make_client(instance):
    """Returns an tacker client."""

    tacker_client = utils.get_client_class(
        API_NAME,
        instance._api_version[API_NAME],
        API_VERSIONS,
    )
    instance.initialize()
    url = instance._url
    url = url.rstrip("/")
    if '1.0' == instance._api_version[API_NAME]:
        client = tacker_client(username=instance._username,
                               tenant_name=instance._tenant_name,
                               password=instance._password,
                               region_name=instance._region_name,
                               auth_url=instance._auth_url,
                               endpoint_url=url,
                               endpoint_type=instance._endpoint_type,
                               token=instance._token,
                               auth_strategy=instance._auth_strategy,
                               insecure=instance._insecure,
                               ca_cert=instance._ca_cert,
                               retries=instance._retries,
                               raise_errors=instance._raise_errors,
                               session=instance._session,
                               auth=instance._auth)
        return client
    else:
        raise exceptions.UnsupportedVersion(
            reason=_("API version %s is not supported") %
            instance._api_version[API_NAME])


def Client(api_version, *args, **kwargs):
    """Return an tacker client.

    :param api_version: only 1.0 is supported now
    """
    tacker_client = utils.get_client_class(
        API_NAME,
        api_version,
        API_VERSIONS,
    )
    return tacker_client(*args, **kwargs)



# Copyright (C) 2022 Fujitsu
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

from osc_lib.command import command
from osc_lib import utils

from tackerclient.i18n import _
from tackerclient.osc import sdk_utils
from tackerclient.osc import utils as tacker_osc_utils

LOG = logging.getLogger(__name__)

_ATTR_MAP = (
    ('id', 'ID', tacker_osc_utils.LIST_BOTH),
    ('managedObjectId', 'Managed Object Id', tacker_osc_utils.LIST_BOTH),
    ('ackState', 'Ack State', tacker_osc_utils.LIST_BOTH),
    ('eventType', 'Event Type', tacker_osc_utils.LIST_BOTH),
    ('perceivedSeverity', 'Perceived Severity', tacker_osc_utils.LIST_BOTH),
    ('probableCause', 'Probable Cause', tacker_osc_utils.LIST_BOTH)
)

_FORMATTERS = {
    'vnfcInstanceIds': tacker_osc_utils.FormatComplexDataColumn,
    'rootCauseFaultyResource': tacker_osc_utils.FormatComplexDataColumn,
    'correlatedAlarmIds': tacker_osc_utils.FormatComplexDataColumn,
    'faultDetails': tacker_osc_utils.FormatComplexDataColumn,
    '_links': tacker_osc_utils.FormatComplexDataColumn
}

_MIXED_CASE_FIELDS = (
    'managedObjectId', 'rootCauseFaultyResource', 'vnfcInstanceIds',
    'alarmRaisedTime', 'alarmChangedTime', 'alarmClearedTime',
    'alarmAcknowledgedTime', 'ackState', 'perceivedSeverity', 'eventTime',
    'eventType', 'faultType', 'probableCause', 'isRootCause',
    'correlatedAlarmIds', 'faultDetails'
)

_VNF_FM_ALARM_ID = 'vnf_fm_alarm_id'


def _get_columns(vnffm_alarm_obj, action=None):
    if action == 'update':
        column_map = {
            'ackState': 'Ack State'
        }
    else:
        column_map = {
            'id': 'ID',
            'managedObjectId': 'Managed Object Id',
            'ackState': 'Ack State',
            'perceivedSeverity': 'Perceived Severity',
            'eventType': 'Event Type',
            'probableCause': 'Probable Cause'
        }

    if action == 'show':
        column_map.update({
            'vnfcInstanceIds': 'Vnfc Instance Ids',
            'rootCauseFaultyResource': 'Root Cause Faulty Resource',
            'alarmRaisedTime': 'Alarm Raised Time',
            'alarmChangedTime': 'Alarm Changed Time',
            'alarmClearedTime': 'Alarm Cleared Time',
            'alarmAcknowledgedTime': 'Alarm Acknowledged Time',
            'eventTime': 'Event Time',
            'faultType': 'Fault Type',
            'isRootCause': 'Is Root Cause',
            'correlatedAlarmIds': 'Correlated Alarm Ids',
            'faultDetails': 'Fault Details',
            '_links': 'Links'
        })

    return sdk_utils.get_osc_show_columns_for_sdk_resource(
        vnffm_alarm_obj, column_map)


class ListVnfFmAlarm(command.Lister):
    _description = _("List VNF FM alarms")

    def get_parser(self, prog_name):
        LOG.debug('get_parser(%s)', prog_name)
        parser = super(ListVnfFmAlarm, self).get_parser(prog_name)
        parser.add_argument(
            "--filter",
            metavar="<filter>",
            help=_("Attribute-based-filtering parameters"),
        )
        return parser

    def take_action(self, parsed_args):
        _params = {}

        if parsed_args.filter:
            _params['filter'] = parsed_args.filter

        client = self.app.client_manager.tackerclient
        data = client.list_vnf_fm_alarms(**_params)
        headers, columns = tacker_osc_utils.get_column_definitions(
            _ATTR_MAP, long_listing=True)
        return (headers,
                (utils.get_dict_properties(
                    s, columns, formatters=_FORMATTERS,
                    mixed_case_fields=_MIXED_CASE_FIELDS,
                ) for s in data['vnf_fm_alarms']))


class ShowVnfFmAlarm(command.ShowOne):
    _description = _("Display VNF FM alarm details")

    def get_parser(self, prog_name):
        parser = super(ShowVnfFmAlarm, self).get_parser(prog_name)
        parser.add_argument(
            _VNF_FM_ALARM_ID,
            metavar="<vnf-fm-alarm-id>",
            help=_("VNF FM alarm ID to display"))
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.tackerclient
        obj = client.show_vnf_fm_alarm(parsed_args.vnf_fm_alarm_id)
        display_columns, columns = _get_columns(obj, action='show')
        data = utils.get_item_properties(
            sdk_utils.DictModel(obj), columns,
            mixed_case_fields=_MIXED_CASE_FIELDS,
            formatters=_FORMATTERS)
        return (display_columns, data)


class UpdateVnfFmAlarm(command.ShowOne):
    _description = _("Update information about an individual VNF FM alarm")

    def get_parser(self, prog_name):
        LOG.debug('get_parser(%s)', prog_name)
        parser = super(UpdateVnfFmAlarm, self).get_parser(prog_name)
        parser.add_argument(
            _VNF_FM_ALARM_ID,
            metavar="<vnf-fm-alarm-id>",
            help=_("VNF FM alarm ID to update.")
        )
        update_require_parameters = parser.add_argument_group(
            "require arguments"
        )
        update_require_parameters.add_argument(
            "--ack-state",
            metavar="<ack-state>",
            choices=['ACKNOWLEDGED', 'UNACKNOWLEDGED'],
            help=_("Ask state can be 'ACKNOWLEDGED' or 'UNACKNOWLEDGED'."))
        return parser

    def args2body(self, parsed_args):
        body = {'ackState': parsed_args.ack_state}
        return body

    def take_action(self, parsed_args):
        client = self.app.client_manager.tackerclient
        updated_values = client.update_vnf_fm_alarm(
            parsed_args.vnf_fm_alarm_id, self.args2body(parsed_args))
        display_columns, columns = _get_columns(
            updated_values, action='update')
        data = utils.get_item_properties(
            sdk_utils.DictModel(updated_values), columns,
            mixed_case_fields=_MIXED_CASE_FIELDS,
            formatters=_FORMATTERS)
        return (display_columns, data)



# Copyright 2012 NEC Corporation
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import copy
import json
from unittest import mock
import uuid

from keystoneclient import exceptions as k_exceptions
import requests
import testtools

from tackerclient import client
from tackerclient.common import exceptions


USERNAME = 'testuser'
USER_ID = 'testuser_id'
TENANT_NAME = 'testtenant'
TENANT_ID = 'testtenant_id'
PASSWORD = 'password'
AUTH_URL = 'authurl'
ENDPOINT_URL = 'localurl'
ENDPOINT_OVERRIDE = 'otherurl'
TOKEN = 'tokentoken'
REGION = 'RegionTest'
NOAUTH = 'noauth'

KS_TOKEN_RESULT = {
    'access': {
        'token': {'id': TOKEN,
                  'expires': '2012-08-11T07:49:01Z',
                  'tenant': {'id': str(uuid.uuid1())}},
        'user': {'id': str(uuid.uuid1())},
        'serviceCatalog': [
            {'endpoints_links': [],
             'endpoints': [{'adminURL': ENDPOINT_URL,
                            'internalURL': ENDPOINT_URL,
                            'publicURL': ENDPOINT_URL,
                            'region': REGION}],
             'type': 'nfv-orchestration',
             'name': 'Tacker Service'}
        ]
    }
}

ENDPOINTS_RESULT = {
    'endpoints': [{
        'type': 'nfv-orchestration',
        'name': 'Tacker Service',
        'region': REGION,
        'adminURL': ENDPOINT_URL,
        'internalURL': ENDPOINT_URL,
        'publicURL': ENDPOINT_URL
    }]
}


def get_response(status_code, headers=None):
    response = mock.Mock().CreateMock(requests.Response)
    response.headers = headers or {}
    response.status_code = status_code
    return response


resp_200 = get_response(200)
resp_401 = get_response(401)
headers = {'X-Auth-Token': '',
           'User-Agent': 'python-tackerclient'}
expected_headers = {'X-Auth-Token': TOKEN,
                    'User-Agent': 'python-tackerclient'}
agent_header = {'User-Agent': 'python-tackerclient'}


class CLITestAuthNoAuth(testtools.TestCase):

    def setUp(self):
        """Prepare the test environment."""
        super(CLITestAuthNoAuth, self).setUp()
        self.client = client.HTTPClient(username=USERNAME,
                                        tenant_name=TENANT_NAME,
                                        password=PASSWORD,
                                        endpoint_url=ENDPOINT_URL,
                                        auth_strategy=NOAUTH,
                                        region_name=REGION)
        self.addCleanup(mock.patch.stopall)

    @mock.patch('tackerclient.client.HTTPClient.request')
    def test_get_noauth(self, mock_request):

        mock_request.return_value = (resp_200, '')
        self.client.do_request('/resource', 'GET',
                               headers=headers)
        mock_request.assert_called_once_with(
            ENDPOINT_URL + '/resource',
            'GET',
            headers=headers,
            content_type=None)
        self.assertEqual(self.client.endpoint_url, ENDPOINT_URL)


class CLITestAuthKeystone(testtools.TestCase):

    # Auth Body expected
    auth_body = ('{"auth": {"tenantName": "testtenant", '
                 '"passwordCredentials": '
                 '{"username": "testuser", "password": "password"}}}')

    def setUp(self):
        """Prepare the test environment."""
        super(CLITestAuthKeystone, self).setUp()
        self.client = client.HTTPClient(username=USERNAME,
                                        tenant_name=TENANT_NAME,
                                        password=PASSWORD,
                                        auth_url=AUTH_URL,
                                        region_name=REGION)
        self.addCleanup(mock.patch.stopall)

    def test_reused_token_get_auth_info(self):
        """Test Client.get_auth_info().

        Test that Client.get_auth_info() works even if client was
        instantiated with predefined token.
        """
        client_ = client.HTTPClient(username=USERNAME,
                                    tenant_name=TENANT_NAME,
                                    token=TOKEN,
                                    password=PASSWORD,
                                    auth_url=AUTH_URL,
                                    region_name=REGION)
        expected = {'auth_token': TOKEN,
                    'auth_tenant_id': None,
                    'auth_user_id': None,
                    'endpoint_url': self.client.endpoint_url}
        self.assertEqual(client_.get_auth_info(), expected)

    @mock.patch('tackerclient.client.HTTPClient.request')
    def test_get_token(self, mock_request):

        mock_request.return_value = (resp_200, json.dumps(KS_TOKEN_RESULT))
        self.client.do_request('/resource', 'GET')
        mock_request.assert_called_with(
            ENDPOINT_URL + '/resource', 'GET',
            headers=expected_headers, content_type=None)
        self.assertEqual(self.client.endpoint_url, ENDPOINT_URL)
        self.assertEqual(self.client.auth_token, TOKEN)

    @mock.patch('tackerclient.client.HTTPClient.request')
    def test_refresh_token(self, mock_request):

        self.client.auth_token = TOKEN
        self.client.endpoint_url = ENDPOINT_URL

        # If a token is expired, tacker server returns 401
        mock_request.return_value = (resp_401, '')
        self.assertRaises(exceptions.Unauthorized,
                          self.client.do_request,
                          '/resource',
                          'GET')

        mock_request.return_value = (resp_200, json.dumps(KS_TOKEN_RESULT))
        self.client.do_request('/resource', 'GET')
        mock_request.assert_called_with(
            ENDPOINT_URL + '/resource', 'GET',
            headers=expected_headers, content_type=None)

    @mock.patch('tackerclient.client.HTTPClient.request')
    def test_refresh_token_no_auth_url(self, mock_request):

        self.client.auth_url = None

        self.client.auth_token = TOKEN
        self.client.endpoint_url = ENDPOINT_URL

        # If a token is expired, tacker server returns 401
        mock_request.return_value = (resp_401, '')
        self.assertRaises(exceptions.NoAuthURLProvided,
                          self.client.do_request,
                          '/resource',
                          'GET')
        expected_url = ENDPOINT_URL + '/resource'
        mock_request.assert_called_with(expected_url, 'GET',
                                        headers=expected_headers,
                                        content_type=None)

    def test_get_endpoint_url_with_invalid_auth_url(self):
        # Handle the case when auth_url is not provided
        self.client.auth_url = None
        self.assertRaises(exceptions.NoAuthURLProvided,
                          self.client._get_endpoint_url)

    @mock.patch('tackerclient.client.HTTPClient.request')
    def test_get_endpoint_url(self, mock_request):

        self.client.auth_token = TOKEN

        mock_request.return_value = (resp_200, json.dumps(ENDPOINTS_RESULT))
        self.client.do_request('/resource', 'GET')
        mock_request.assert_called_with(
            ENDPOINT_URL + '/resource', 'GET',
            headers=expected_headers, content_type=None)

        mock_request.return_value = (resp_200, '')
        self.client.do_request('/resource', 'GET',
                               headers=headers)
        mock_request.assert_called_with(
            ENDPOINT_URL + '/resource', 'GET',
            headers=headers, content_type=None)

    @mock.patch('tackerclient.client.HTTPClient.request')
    def test_use_given_endpoint_url(self, mock_request):
        self.client = client.HTTPClient(
            username=USERNAME, tenant_name=TENANT_NAME, password=PASSWORD,
            auth_url=AUTH_URL, region_name=REGION,
            endpoint_url=ENDPOINT_OVERRIDE)
        self.assertEqual(self.client.endpoint_url, ENDPOINT_OVERRIDE)

        self.client.auth_token = TOKEN
        mock_request.return_value = (resp_200, '')

        self.client.do_request('/resource', 'GET',
                               headers=headers)
        mock_request.assert_called_with(
            ENDPOINT_OVERRIDE + '/resource', 'GET',
            headers=headers, content_type=None)
        self.assertEqual(self.client.endpoint_url, ENDPOINT_OVERRIDE)

    @mock.patch('tackerclient.client.HTTPClient.request')
    def test_get_endpoint_url_other(self, mock_request):
        self.client = client.HTTPClient(
            username=USERNAME, tenant_name=TENANT_NAME, password=PASSWORD,
            auth_url=AUTH_URL, region_name=REGION, endpoint_type='otherURL')

        self.client.auth_token = TOKEN
        mock_request.return_value = (resp_200, json.dumps(ENDPOINTS_RESULT))
        self.assertRaises(exceptions.EndpointTypeNotFound,
                          self.client.do_request,
                          '/resource',
                          'GET')
        expected_url = AUTH_URL + '/tokens/%s/endpoints' % TOKEN
        headers = {'User-Agent': 'python-tackerclient'}
        mock_request.assert_called_with(expected_url, 'GET',
                                        headers=headers)

    @mock.patch('tackerclient.client.HTTPClient.request')
    def test_get_endpoint_url_failed(self, mock_request):

        self.client.auth_token = TOKEN
        self.client.auth_url = AUTH_URL + '/tokens/%s/endpoints' % TOKEN

        mock_request.return_value = (resp_401, '')
        self.assertRaises(exceptions.Unauthorized,
                          self.client.do_request,
                          '/resource',
                          'GET')

    def test_endpoint_type(self):
        resources = copy.deepcopy(KS_TOKEN_RESULT)
        endpoints = resources['access']['serviceCatalog'][0]['endpoints'][0]
        endpoints['internalURL'] = 'internal'
        endpoints['adminURL'] = 'admin'
        endpoints['publicURL'] = 'public'

        # Test default behavior is to choose public.
        self.client = client.HTTPClient(
            username=USERNAME, tenant_name=TENANT_NAME, password=PASSWORD,
            auth_url=AUTH_URL, region_name=REGION)

        self.client._extract_service_catalog(resources)
        self.assertEqual(self.client.endpoint_url, 'public')

        # Test admin url
        self.client = client.HTTPClient(
            username=USERNAME, tenant_name=TENANT_NAME, password=PASSWORD,
            auth_url=AUTH_URL, region_name=REGION, endpoint_type='adminURL')

        self.client._extract_service_catalog(resources)
        self.assertEqual(self.client.endpoint_url, 'admin')

        # Test public url
        self.client = client.HTTPClient(
            username=USERNAME, tenant_name=TENANT_NAME, password=PASSWORD,
            auth_url=AUTH_URL, region_name=REGION, endpoint_type='publicURL')

        self.client._extract_service_catalog(resources)
        self.assertEqual(self.client.endpoint_url, 'public')

        # Test internal url
        self.client = client.HTTPClient(
            username=USERNAME, tenant_name=TENANT_NAME, password=PASSWORD,
            auth_url=AUTH_URL, region_name=REGION, endpoint_type='internalURL')

        self.client._extract_service_catalog(resources)
        self.assertEqual(self.client.endpoint_url, 'internal')

        # Test url that isn't found in the service catalog
        self.client = client.HTTPClient(
            username=USERNAME, tenant_name=TENANT_NAME, password=PASSWORD,
            auth_url=AUTH_URL, region_name=REGION, endpoint_type='privateURL')

        self.assertRaises(k_exceptions.EndpointNotFound,
                          self.client._extract_service_catalog,
                          resources)

    @mock.patch('tackerclient.client.HTTPClient.request')
    @mock.patch('tackerclient.common.utils.http_log_req')
    def test_strip_credentials_from_log(self, mock_http_log_req,
                                        mock_request,):

        body = ('{"auth": {"tenantId": "testtenant_id",'
                '"passwordCredentials": {"password": "password",'
                '"userId": "testuser_id"}}}')
        expected_body = ('{"auth": {"tenantId": "testtenant_id",'
                         '"REDACTEDCredentials": {"REDACTED": "REDACTED",'
                         '"userId": "testuser_id"}}}')
        _headers = {'headers': expected_headers, 'body': expected_body,
                    'content_type': None}

        mock_request.return_value = (resp_200, json.dumps(KS_TOKEN_RESULT))
        self.client.do_request('/resource', 'GET', body=body)

        args, kwargs = mock_http_log_req.call_args
        # Check that credentials are stripped while logging.
        self.assertEqual(_headers, args[2])


class CLITestAuthKeystoneWithId(CLITestAuthKeystone):

    # Auth Body expected
    auth_body = ('{"auth": {"passwordCredentials": '
                 '{"password": "password", "userId": "testuser_id"}, '
                 '"tenantId": "testtenant_id"}}')

    def setUp(self):
        """Prepare the test environment."""
        super(CLITestAuthKeystoneWithId, self).setUp()
        self.client = client.HTTPClient(user_id=USER_ID,
                                        tenant_id=TENANT_ID,
                                        password=PASSWORD,
                                        auth_url=AUTH_URL,
                                        region_name=REGION)


class CLITestAuthKeystoneWithIdandName(CLITestAuthKeystone):

    # Auth Body expected
    auth_body = ('{"auth": {"passwordCredentials": '
                 '{"password": "password", "userId": "testuser_id"}, '
                 '"tenantId": "testtenant_id"}}')

    def setUp(self):
        """Prepare the test environment."""
        super(CLITestAuthKeystoneWithIdandName, self).setUp()
        self.client = client.HTTPClient(username=USERNAME,
                                        user_id=USER_ID,
                                        tenant_id=TENANT_ID,
                                        tenant_name=TENANT_NAME,
                                        password=PASSWORD,
                                        auth_url=AUTH_URL,
                                        region_name=REGION)
